from typing import Any, List, Mapping, Sequence, Tuple

import pyro
import pyro.distributions as distributions
import torch
import torch.distributions.constraints as constraints
from pyro.distributions.util import broadcast_shape
from pyro.ops.indexing import Vindex

from pearl.common import NodeValueType, SamplingMode
from pearl.nodes.bayesnetnode import Node, NodeWithCategoricalParents


class CategoricalNodeWithDirichletPrior(NodeWithCategoricalParents):
    """Categorically distributed node with Dirichlet prior."""

    def __init__(
        self,
        name: str,
        domain: Sequence[str],
        plates: List[str],
        parents: Sequence[Node],
        device: torch.device = torch.device("cpu"),
        observed: bool = True,
        prior_params: Mapping[str, torch.Tensor] = None,
    ) -> None:

        if len(domain) < 2:
            raise ValueError(
                f"Domain of categorical node {name} must include at-least two categories."
            )
        self.domain = domain

        super().__init__(
            name,
            plates,
            parents,
            observed=observed,
            device=device,
            prior_params=prior_params,
        )
        self.guide_MAP_cpd = None

    @property
    def domain_size(self):
        return len(self.domain)

    def value_type(self) -> NodeValueType:
        return NodeValueType.CATEGORICAL

    def dist(self, probs) -> distributions.Distribution:
        return distributions.Categorical(probs)

    def get_default_prior_params(self):
        return {
            "alpha": torch.ones(
                self.parent_domain_sizes + (self.domain_size,), device=self.device
            )
        }

    def validate_prior_params(self):
        super().validate_prior_params()
        assert self.prior_params["alpha"].shape == self.parent_domain_sizes + (
            self.domain_size,
        )
        assert self.prior_params["alpha"].gt(0).all()

    def sample_model_cpd(self) -> torch.Tensor:
        with pyro.plate_stack(f"{self.name}/cpd/plate", self.parent_domain_sizes):
            self.model_cpd = pyro.sample(
                f"{self.name}/cpd", distributions.Dirichlet(self.prior_params["alpha"])
            )
        if self.model_csi is not None:
            self.model_cpd = self.model_csi.map(self.model_cpd)
        return self.model_cpd

    def sample_guide_cpd(self) -> torch.Tensor:
        self.guide_alpha = pyro.param(
            f"{self.name}/guide/alpha",
            self.prior_params["alpha"],
            constraint=constraints.positive,
        )
        with pyro.plate_stack(f"{self.name}/cpd/plate", self.parent_domain_sizes):
            self.guide_cpd = pyro.sample(
                f"{self.name}/cpd", distributions.Dirichlet(self.guide_alpha)
            )
        if self.guide_csi is not None:
            self.guide_cpd = self.guide_csi.map(self.guide_cpd)
        return self.guide_cpd

    def MAP_cpd(self) -> torch.Tensor:
        """Return the MAP estimate of the model CPD."""
        self.guide_MAP_cpd = distributions.Dirichlet(self.guide_alpha).mean
        if self.guide_csi is not None:
            self.guide_MAP_cpd = self.guide_csi.map(self.guide_MAP_cpd)
        return self.guide_MAP_cpd

    def overwrite_param_store(self) -> None:
        param_store = pyro.get_param_store()
        assert self.guide_alpha is not None, "Guide hyperparameters don't exist"
        param_store["/".join((self.name, "guide", "alpha"))] = self.guide_alpha

    def to_yaml_encoding(self) -> Mapping[str, Any]:
        self.sample_guide_cpd()  # to ensure that guide parameters are set
        yaml_encoding = dict(super().to_yaml_encoding())
        yaml_encoding["domain"] = list(self.domain)
        yaml_encoding["prior_params"] = {"alpha": self.guide_alpha.tolist()}
        return yaml_encoding

    @classmethod
    def process_yaml_spec(
        cls, bayesnet, yaml_encoding: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        kwargs = super().process_yaml_spec(bayesnet, yaml_encoding)
        kwargs["domain"] = yaml_encoding["domain"]
        if "prior_params" in yaml_encoding:
            kwargs["prior_params"] = {
                "alpha": torch.tensor(
                    yaml_encoding["prior_params"]["alpha"], device=bayesnet.device
                )
            }
        return kwargs


class GeneralizedLinearNode(Node):
    """
    Node which implements the generlized linear model for sampling a
    categorical random variable conditioned on a mix of categorical
    and continuous parents.

    This node requires atleast one continuous parent.  For each
    combination of discrete parents, the random variable is sampled
    from a categorical distribution that is computed using a softmax
    function applied to a linear combination of the continuous
    parents.

    If the discrete parents have domains of size D1,...,Dk, the
    number of categories are M, and the number of continuous parents
    are C, then it uses 'bias' and 'weight' tensors of shape [D1, ...,
    Dk, M] and [D1, ..., Dk, M, C].  Instead of using point estimates,
    these tensors are sampled from normal priors over them.
    """

    def __init__(
        self,
        name: str,
        domain: Sequence[str],
        plates: List[str],
        parents: Sequence[Node],
        device: torch.device = torch.device("cpu"),
        prior_params: Mapping[str, torch.Tensor] = None,
        observed: bool = True,
    ) -> None:
        self.domain = domain

        # record positions of discrete and continuous parents
        discrete_parent_indices = []
        continuous_parent_indices = []
        for i, p in enumerate(parents):
            if p.value_type() == NodeValueType.CATEGORICAL:
                discrete_parent_indices.append(i)
            elif p.value_type() == NodeValueType.CONTINUOUS:
                continuous_parent_indices.append(i)
            else:
                raise ValueError(
                    f"unsupported parent of type {p.value_type()} for node {name}"
                )
        assert (
            len(continuous_parent_indices) > 0
        ), f"GeneralizedLinearNode {name} requires atleast one continuous parent"
        self.discrete_parent_indices = discrete_parent_indices
        self.continuous_parent_indices = continuous_parent_indices

        # get domain sizes of discrete parents
        discrete_parent_domain_sizes = []
        for i in self.discrete_parent_indices:
            discrete_parent_domain_sizes.append(parents[i].domain_size)
        self.discrete_parent_domain_sizes = torch.Size(discrete_parent_domain_sizes)

        super().__init__(
            name=name,
            plates=plates,
            parents=parents,
            device=device,
            prior_params=prior_params,
            observed=observed,
        )

    @property
    def domain_size(self):
        return len(self.domain)

    def value_type(self) -> NodeValueType:
        return NodeValueType.CATEGORICAL

    def dist(self, probs) -> distributions.Distribution:
        return distributions.Categorical(probs)

    def get_default_prior_params(self):
        return {
            "bias_mean": torch.zeros(
                *self.discrete_parent_domain_sizes, len(self.domain)
            ),
            "bias_scale": torch.ones(
                *self.discrete_parent_domain_sizes, len(self.domain)
            ),
            "weights_mean": torch.zeros(
                *self.discrete_parent_domain_sizes,
                len(self.continuous_parent_indices),
                len(self.domain),
            ),
            "weights_scale": torch.ones(
                *self.discrete_parent_domain_sizes,
                len(self.continuous_parent_indices),
                len(self.domain),
            ),
        }

    def validate_prior_params(self):
        super().validate_prior_params()
        assert self.prior_params["bias_mean"].shape == (
            *self.discrete_parent_domain_sizes,
            len(self.domain),
        )
        assert self.prior_params["bias_scale"].shape == (
            *self.discrete_parent_domain_sizes,
            len(self.domain),
        )
        assert self.prior_params["weights_mean"].shape == (
            *self.discrete_parent_domain_sizes,
            len(self.continuous_parent_indices),
            len(self.domain),
        )
        assert self.prior_params["weights_scale"].shape == (
            *self.discrete_parent_domain_sizes,
            len(self.continuous_parent_indices),
            len(self.domain),
        )

    def sample_model_cpd(self) -> Tuple[torch.Tensor, torch.Tensor]:
        with pyro.plate_stack(
            f"{self.name}/cpd/bias/plate", self.prior_params["bias_mean"].shape
        ):
            bias = pyro.sample(
                f"{self.name}/bias",
                distributions.Normal(
                    self.prior_params["bias_mean"],
                    self.prior_params["bias_scale"],
                ),
            )

        with pyro.plate_stack(
            f"{self.name}/cpd/weights/plate",
            self.prior_params["weights_mean"].shape,
        ):
            weights = pyro.sample(
                f"{self.name}/weights",
                distributions.Normal(
                    self.prior_params["weights_mean"],
                    self.prior_params["weights_scale"],
                ),
            )
        self.model_cpd = (weights, bias)
        return self.model_cpd

    def sample_guide_cpd(self) -> Tuple[torch.Tensor, torch.Tensor]:
        self.guide_bias_mean = pyro.param(
            f"{self.name}/guide/bias_mean",
            self.prior_params["bias_mean"],
        )
        self.guide_bias_scale = pyro.param(
            f"{self.name}/guide/bias_scale",
            self.prior_params["bias_scale"],
            constraint=constraints.positive,
        )
        self.guide_weights_mean = pyro.param(
            f"{self.name}/guide/weights_mean",
            self.prior_params["weights_mean"],
        )
        self.guide_weights_scale = pyro.param(
            f"{self.name}/guide/weights_scale",
            self.prior_params["weights_scale"],
            constraint=constraints.positive,
        )
        with pyro.plate_stack(
            f"{self.name}/cpd/bias/plate", self.guide_bias_mean.shape
        ):
            bias = pyro.sample(
                f"{self.name}/bias",
                distributions.Normal(
                    self.guide_bias_mean,
                    self.guide_bias_scale,
                ),
            )
        with pyro.plate_stack(
            f"{self.name}/cpd/weights/plate", self.guide_weights_mean.shape
        ):
            weights = pyro.sample(
                f"{self.name}/weights",
                distributions.Normal(
                    self.guide_weights_mean,
                    self.guide_weights_scale,
                ),
            )
        self.guide_cpd = (weights, bias)
        return self.guide_cpd

    def MAP_cpd(self) -> Tuple[torch.Tensor, torch.Tensor]:
        return (self.guide_weights_mean, self.guide_bias_mean)

    def _index_cpd(
        self, cpd: torch.Tensor, *parent_values: torch.Tensor
    ) -> List[torch.Tensor]:
        """
        Index the conditional probability distribution of the node
        using parent values and return it.

        :param cpd: tensor representing the CPD of the node
        :param parent_values: tensors containing values of the parents
            of the node
        """
        assert len(parent_values) == len(self.discrete_parent_indices)
        if len(parent_values) > 0:
            t_parent_values = tuple(p.long() for p in parent_values)
            return Vindex(cpd)[t_parent_values]
        else:
            return cpd

    def get_sample_dist(
        self, sampling_mode: SamplingMode, *parent_values: torch.Tensor
    ) -> distributions.Distribution:
        if sampling_mode == SamplingMode.MODEL:
            weights, bias = self.sample_model_cpd()
        elif sampling_mode == SamplingMode.GUIDE:
            weights, bias = self.sample_guide_cpd()
        elif sampling_mode == SamplingMode.MAP_POSTERIOR:
            weights, bias = self.MAP_cpd()
        else:
            raise ValueError(f"uncrecognized sampling mode- {sampling_mode}")

        # index the cpd based on values of discrete parents
        discrete_parent_values = [
            parent_values[i] for i in self.discrete_parent_indices
        ]
        weights = self._index_cpd(weights, *discrete_parent_values)
        bias = self._index_cpd(bias, *discrete_parent_values)

        # stack continuous parent values after broadcasting them
        continuous_parent_shapes = [
            parent_values[i].shape for i in self.continuous_parent_indices
        ]
        broadcasted_shape = broadcast_shape(*continuous_parent_shapes)
        stacked_continuous_parent_tensors = torch.stack(
            [
                parent_values[i].expand(broadcasted_shape)
                for i in self.continuous_parent_indices
            ],
            dim=-1,
        )
        stacked_continuous_parent_tensors.unsqueeze_(
            -1
        )  # add a dimension for categories
        probs = torch.softmax(
            torch.sum(stacked_continuous_parent_tensors * weights, dim=-2) + bias,
            dim=-1,
        )
        return self.dist(probs)

    def overwrite_param_store(self) -> None:
        param_store = pyro.get_param_store()
        param_store[
            "/".join((self.name, "guide", "weights_mean"))
        ] = self.guide_weights_mean
        param_store[
            "/".join((self.name, "guide", "weights_scale"))
        ] = self.guide_weights_scale
        param_store["/".join((self.name, "guide", "bias_mean"))] = self.guide_bias_mean
        param_store[
            "/".join((self.name, "guide", "bias_scale"))
        ] = self.guide_bias_scale

    def to_yaml_encoding(self) -> Mapping[str, Any]:
        self.sample_guide_cpd()
        yaml_encoding = dict(super().to_yaml_encoding())
        yaml_encoding["domain"] = list(self.domain)
        yaml_encoding["prior_params"] = {
            "weights_mean": self.guide_weights_mean.tolist(),
            "weights_scale": self.guide_weights_scale.tolist(),
            "bias_mean": self.guide_bias_mean.tolist(),
            "bias_scale": self.guide_bias_scale.tolist(),
        }
        return yaml_encoding

    @classmethod
    def process_yaml_spec(
        cls, bayesnet, yaml_encoding: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        kwargs = super().process_yaml_spec(bayesnet, yaml_encoding)
        kwargs["domain"] = yaml_encoding["domain"]
        if "prior_params" in yaml_encoding:
            kwargs["prior_params"] = {
                "weights_mean": torch.tensor(
                    yaml_encoding["prior_params"]["weights_mean"],
                    device=bayesnet.device,
                ),
                "weights_scale": torch.tensor(
                    yaml_encoding["prior_params"]["weights_scale"],
                    device=bayesnet.device,
                ),
                "bias_mean": torch.tensor(
                    yaml_encoding["prior_params"]["bias_mean"], device=bayesnet.device
                ),
                "bias_scale": torch.tensor(
                    yaml_encoding["prior_params"]["bias_scale"], device=bayesnet.device
                ),
            }
        return kwargs



from typing import Any, List, Mapping, Sequence, Tuple

import pyro
import pyro.distributions as distributions
import torch
import torch.distributions.constraints as constraints
from pyro.distributions.util import broadcast_shape
from pyro.ops.indexing import Vindex

from pearl.common import NodeValueType, SamplingMode
from pearl.nodes.bayesnetnode import Node, NodeWithCategoricalParents


class ContinuousNodeWithNormalDistribution(NodeWithCategoricalParents):
    """Node whose continuous value is drawn from a Normal distribution. The prior for the mean of the normal
    distribution is given by a normal distribution itself, specified with "mean_mean" and "mean_scale" prior
    parameters with defaults 0.0 and 1.0, while the prior for the scale of the normal distribution is given by a
    HalfCauchy distribution with scale "scale_scale", 1.0 by default. If any of the prior parameters are provided as
    individual floats then those values will be expanded to tensors matching parent domain sizes.
    """

    def __init__(
        self,
        name: str,
        plates: List[str],
        parents: Sequence[Node],
        device: torch.device = torch.device("cpu"),
        observed: bool = True,
        prior_params: Mapping[str, torch.Tensor] = None,
    ) -> None:

        super().__init__(
            name,
            plates,
            parents,
            observed=observed,
            device=device,
            prior_params=prior_params,
        )
        self.guide_MAP_cpd = None

    def get_default_prior_params(self):
        # The tuple value (1,) catted to the end of parent_domain_sizes ensures
        # that we have parameters for the distribution even if the node has no parents.
        return {
            # Mean of normal prior on mean.
            "mean_mean": torch.zeros(self.parent_domain_sizes, device=self.device),
            # Scale of normal prior on mean.
            "mean_scale": torch.ones(self.parent_domain_sizes, device=self.device),
            # Scale of half-cauchy prior on scale.
            "scale_scale": torch.ones(self.parent_domain_sizes, device=self.device),
        }

    def value_type(self) -> NodeValueType:
        return NodeValueType.CONTINUOUS

    def dist(self, loc_and_scale) -> distributions.Distribution:
        return distributions.Normal(
            loc_and_scale[Ellipsis, 0], loc_and_scale[Ellipsis, 1]
        )

    def validate_prior_params(self):
        super().validate_prior_params()
        for param_name in ["mean_mean", "mean_scale", "scale_scale"]:
            assert self.prior_params[param_name].shape == self.parent_domain_sizes

        for param_name in ["mean_scale", "scale_scale"]:
            assert self.prior_params[param_name].gt(0).all()

    def sample_model_cpd(self) -> None:
        with pyro.plate_stack(f"{self.name}/cpd/plate", self.parent_domain_sizes):
            mean = pyro.sample(
                f"{self.name}/cpd/mean",
                distributions.Normal(
                    self.prior_params["mean_mean"], self.prior_params["mean_scale"]
                ),
            )

            scale = pyro.sample(
                f"{self.name}/cpd/scale",
                distributions.HalfCauchy(self.prior_params["scale_scale"]),
            )

            if self.model_csi is not None:
                mean = self.model_csi.map(mean)
                scale = self.model_csi.map(scale)

            self.model_cpd = torch.stack((mean, scale), dim=-1)
        return self.model_cpd

    def sample_guide_cpd(self) -> None:
        self.guide_mean_mean = pyro.param(
            f"{self.name}/guide/mean_mean", self.prior_params["mean_mean"]
        )
        self.guide_mean_scale = pyro.param(
            f"{self.name}/guide/mean_scale",
            self.prior_params["mean_scale"],
            constraint=constraints.positive,
        )
        self.guide_scale = pyro.param(
            f"{self.name}/guide/scale",
            self.prior_params["scale_scale"],
            constraint=constraints.positive,
        )

        with pyro.plate_stack(f"{self.name}/cpd/plate", self.parent_domain_sizes):
            mean = pyro.sample(
                f"{self.name}/cpd/mean",
                distributions.Normal(self.guide_mean_mean, self.guide_mean_scale),
            )
            scale = pyro.sample(
                f"{self.name}/cpd/scale", distributions.Delta(self.guide_scale)
            )

            if self.guide_csi is not None:
                mean = self.guide_csi.map(mean)
                scale = self.guide_csi.map(scale)

            self.guide_cpd = torch.stack((mean, scale), dim=-1)
        return self.guide_cpd

    def MAP_cpd(self) -> torch.Tensor:
        guide_mean_mean = self.guide_mean_mean
        guide_scale = self.guide_scale
        if self.guide_csi is not None:
            guide_mean_mean = self.guide_csi.map(guide_mean_mean)
            guide_scale = self.guide_csi.map(guide_scale)
        self.guide_MAP_cpd = torch.stack(
            (guide_mean_mean, guide_scale), dim=-1
        ).squeeze()
        return self.guide_MAP_cpd

    def overwrite_param_store(self) -> None:
        param_store = pyro.get_param_store()

        params = (self.guide_mean_mean, self.guide_mean_scale, self.guide_scale)
        assert all(param is not None for param in params)

        param_store["/".join((self.name, "guide", "mean_mean"))] = self.guide_mean_mean
        param_store[
            "/".join((self.name, "guide", "mean_scale"))
        ] = self.guide_mean_scale
        param_store["/".join((self.name, "guide", "scale"))] = self.guide_scale

    def to_yaml_encoding(self) -> Mapping[str, Any]:
        self.sample_guide_cpd()  # to ensure that guide parameters are set
        yaml_encoding = dict(super().to_yaml_encoding())
        yaml_encoding["prior_params"] = {
            "mean_mean": self.guide_mean_mean.tolist(),
            "mean_scale": self.guide_mean_scale.tolist(),
            "scale_scale": self.guide_scale.tolist(),
        }
        return yaml_encoding

    @classmethod
    def process_yaml_spec(
        cls, bayesnet, yaml_encoding: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        kwargs = super().process_yaml_spec(bayesnet, yaml_encoding)
        if "prior_params" in yaml_encoding:
            kwargs["prior_params"] = {
                "mean_mean": torch.tensor(
                    yaml_encoding["prior_params"]["mean_mean"],
                    device=bayesnet.device,
                ),
                "mean_scale": torch.tensor(
                    yaml_encoding["prior_params"]["mean_scale"],
                    device=bayesnet.device,
                ),
                "scale_scale": torch.tensor(
                    yaml_encoding["prior_params"]["scale_scale"],
                    device=bayesnet.device,
                ),
            }
        return kwargs


class ConditionalLinearGaussianNode(Node):
    """
    Node which implements the conditional linear gaussian model for
    sampling a continuous random variable conditioned on a mix of
    categorical and continuous parents.

    This node requires atleast one continuous parent.  For each
    combination of discrete parents, the random variable is sampled
    from a Normal distribution whose mean is the linear combination of
    its continuous parents and a variance that is independent of
    parent values.

    If the discrete parents have domains of size D1, ..., Dk, and
    there are C continuous parents, then it uses 'bias' and 'weight'
    tensors of shape [D1, ..., Dk] and [D1, ..., Dk, C].  Instead of
    using point estimates, these tensors are sampled from normal
    priors over them.  Further the variant is also sampled from a
    half-cauchy prior.
    """

    def __init__(
        self,
        name: str,
        plates: List[str],
        parents: Sequence[Node],
        device: torch.device = torch.device("cpu"),
        prior_params: Mapping[str, torch.Tensor] = None,
        observed: bool = True,
    ) -> None:
        # record positions of discrete and continuous parents
        discrete_parent_indices = []
        continuous_parent_indices = []
        for i, p in enumerate(parents):
            if p.value_type() == NodeValueType.CATEGORICAL:
                discrete_parent_indices.append(i)
            elif p.value_type() == NodeValueType.CONTINUOUS:
                continuous_parent_indices.append(i)
            else:
                raise ValueError(
                    f"unsupported parent of type {p.value_type()} for node {name}"
                )
        assert (
            len(continuous_parent_indices) > 0
        ), f"ConditionalLinearGaussianNode {name} requires atleast one continuous parent"
        self.discrete_parent_indices = discrete_parent_indices
        self.continuous_parent_indices = continuous_parent_indices

        # get domain sizes of discrete parents
        discrete_parent_domain_sizes = []
        for i in self.discrete_parent_indices:
            discrete_parent_domain_sizes.append(parents[i].domain_size)
        self.discrete_parent_domain_sizes = torch.Size(discrete_parent_domain_sizes)

        super().__init__(
            name=name,
            plates=plates,
            parents=parents,
            device=device,
            prior_params=prior_params,
            observed=observed,
        )

    def value_type(self) -> NodeValueType:
        return NodeValueType.CONTINUOUS

    def dist(self, loc, scale) -> distributions.Distribution:
        return distributions.Normal(loc, scale)

    def get_default_prior_params(self):
        return {
            "bias_mean": torch.full(self.discrete_parent_domain_sizes, 0.0),
            "bias_scale": torch.full(self.discrete_parent_domain_sizes, 1.0),
            "weights_mean": torch.zeros(
                *self.discrete_parent_domain_sizes, len(self.continuous_parent_indices)
            ),
            "weights_scale": torch.ones(
                *self.discrete_parent_domain_sizes, len(self.continuous_parent_indices)
            ),
            "scale_scale": torch.full(self.discrete_parent_domain_sizes, 1.0),
        }

    def validate_prior_params(self):
        super().validate_prior_params()
        assert self.prior_params["bias_mean"].shape == (
            *self.discrete_parent_domain_sizes,
        )
        assert self.prior_params["bias_scale"].shape == (
            *self.discrete_parent_domain_sizes,
        )
        assert self.prior_params["weights_mean"].shape == (
            *self.discrete_parent_domain_sizes,
            len(self.continuous_parent_indices),
        )
        assert self.prior_params["weights_scale"].shape == (
            *self.discrete_parent_domain_sizes,
            len(self.continuous_parent_indices),
        )
        assert self.prior_params["scale_scale"].shape == (
            *self.discrete_parent_domain_sizes,
        )

    def sample_model_cpd(self) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        with pyro.plate_stack(
            f"{self.name}/cpd/bias/plate", self.prior_params["bias_mean"].shape
        ):
            bias = pyro.sample(
                f"{self.name}/bias",
                distributions.Normal(
                    self.prior_params["bias_mean"],
                    self.prior_params["bias_scale"],
                ),
            )
        with pyro.plate_stack(
            f"{self.name}/cpd/weights/plate", self.prior_params["weights_mean"].shape
        ):
            weights = pyro.sample(
                f"{self.name}/weights",
                distributions.Normal(
                    self.prior_params["weights_mean"],
                    self.prior_params["weights_scale"],
                ),
            )
        with pyro.plate_stack(
            f"{self.name}/cpd/scale/plate", self.prior_params["scale_scale"].shape
        ):
            scale = pyro.sample(
                f"{self.name}/scale",
                distributions.HalfCauchy(self.prior_params["scale_scale"]),
            )
        self.model_cpd = (weights, bias, scale)
        return self.model_cpd

    def sample_guide_cpd(self) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        self.guide_bias_mean = pyro.param(
            f"{self.name}/guide/bias_mean",
            self.prior_params["bias_mean"],
        )
        self.guide_bias_scale = pyro.param(
            f"{self.name}/guide/bias_scale",
            self.prior_params["bias_scale"],
            constraint=constraints.positive,
        )
        self.guide_weights_mean = pyro.param(
            f"{self.name}/guide/weights_mean",
            self.prior_params["weights_mean"],
        )
        self.guide_weights_scale = pyro.param(
            f"{self.name}/guide/weights_scale",
            self.prior_params["weights_scale"],
            constraint=constraints.positive,
        )
        self.guide_scale_scale = pyro.param(
            f"{self.name}/guide/scale_scale",
            self.prior_params["scale_scale"],
            constraint=constraints.positive,
        )
        with pyro.plate_stack(
            f"{self.name}/cpd/bias/plate", self.guide_bias_mean.shape
        ):
            bias = pyro.sample(
                f"{self.name}/bias",
                distributions.Normal(
                    self.guide_bias_mean,
                    self.guide_bias_scale,
                ),
            )
        with pyro.plate_stack(
            f"{self.name}/cpd/weights/plate", self.guide_weights_mean.shape
        ):
            weights = pyro.sample(
                f"{self.name}/weights",
                distributions.Normal(
                    self.guide_weights_mean,
                    self.guide_weights_scale,
                ),
            )
        with pyro.plate_stack(
            f"{self.name}/cpd/scale/plate", self.guide_scale_scale.shape
        ):
            scale = pyro.sample(
                f"{self.name}/scale",
                distributions.Delta(self.guide_scale_scale),
            )
        self.guide_cpd = (weights, bias, scale)
        return self.guide_cpd

    def MAP_cpd(self) -> Tuple[torch.Tensor, torch.Tensor]:
        return (self.guide_weights_mean, self.guide_bias_mean, self.guide_scale_scale)

    def _index_cpd(
        self, cpd: torch.Tensor, *parent_values: torch.Tensor
    ) -> List[torch.Tensor]:
        """
        Index the conditional probability distribution of the node
        using parent values and return it.

        :param cpd: tensor representing the CPD of the node
        :param parent_values: tensors containing values of the parents
            of the node
        """
        assert len(parent_values) == len(self.discrete_parent_indices)
        if len(parent_values) > 0:
            t_parent_values = tuple(p.long() for p in parent_values)
            return Vindex(cpd)[t_parent_values]
        else:
            return cpd

    def get_sample_dist(
        self, sampling_mode: SamplingMode, *parent_values: torch.Tensor
    ) -> distributions.Distribution:
        if sampling_mode == SamplingMode.MODEL:
            weights, bias, scale = self.sample_model_cpd()
        elif sampling_mode == SamplingMode.GUIDE:
            weights, bias, scale = self.sample_guide_cpd()
        elif sampling_mode == SamplingMode.MAP_POSTERIOR:
            weights, bias, scale = self.MAP_cpd()
        else:
            raise ValueError(f"unrecognized sampling mode- {sampling_mode}")

        # index the cpd based on values of discrete parents
        discrete_parent_values = [
            parent_values[i] for i in self.discrete_parent_indices
        ]
        weights = self._index_cpd(weights, *discrete_parent_values)
        bias = self._index_cpd(bias, *discrete_parent_values)
        scale = self._index_cpd(scale, *discrete_parent_values)

        continuous_parent_shapes = [
            parent_values[i].shape for i in self.continuous_parent_indices
        ]
        broadcasted_shape = broadcast_shape(*continuous_parent_shapes)

        stacked_continuous_parent_tensors = torch.stack(
            [
                parent_values[i].expand(broadcasted_shape)
                for i in self.continuous_parent_indices
            ],
            dim=-1,
        )
        mean = torch.sum(stacked_continuous_parent_tensors * weights, dim=-1) + bias
        return self.dist(mean, scale)

    def overwrite_param_store(self) -> None:
        param_store = pyro.get_param_store()
        param_store[
            "/".join((self.name, "guide", "weights_mean")),
        ] = self.guide_weights_mean
        param_store[
            "/".join((self.name, "guide", "weights_scale")),
        ] = self.guide_weights_scale
        param_store["/".join((self.name, "guide", "bias_mean"))] = self.guide_bias_mean
        param_store[
            "/".join((self.name, "guide", "bias_scale"))
        ] = self.guide_bias_scale
        param_store[
            "/".join((self.name, "guide", "scale_scale"))
        ] = self.guide_scale_scale

    def to_yaml_encoding(self) -> Mapping[str, Any]:
        self.sample_guide_cpd()
        yaml_encoding = dict(super().to_yaml_encoding())
        yaml_encoding["prior_params"] = {
            "weights_mean": self.guide_weights_mean.tolist(),
            "weights_scale": self.guide_weights_scale.tolist(),
            "bias_mean": self.guide_bias_mean.tolist(),
            "bias_scale": self.guide_bias_scale.tolist(),
            "scale_scale": self.guide_scale_scale.tolist(),
        }
        return yaml_encoding

    @classmethod
    def process_yaml_spec(
        cls, bayesnet, yaml_encoding: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        kwargs = super().process_yaml_spec(bayesnet, yaml_encoding)
        if "prior_params" in yaml_encoding:
            kwargs["prior_params"] = {
                "weights_mean": torch.tensor(
                    yaml_encoding["prior_params"]["weights_mean"],
                    device=bayesnet.device,
                ),
                "weights_scale": torch.tensor(
                    yaml_encoding["prior_params"]["weights_scale"],
                    device=bayesnet.device,
                ),
                "bias_mean": torch.tensor(
                    yaml_encoding["prior_params"]["bias_mean"],
                    device=bayesnet.device,
                ),
                "bias_scale": torch.tensor(
                    yaml_encoding["prior_params"]["bias_scale"],
                    device=bayesnet.device,
                ),
                "scale_scale": torch.tensor(
                    yaml_encoding["prior_params"]["scale_scale"],
                    device=bayesnet.device,
                ),
            }
        return kwargs



import copy
from collections.abc import Iterable
from functools import reduce
from itertools import accumulate
from typing import List, Mapping, NamedTuple, Optional, Sequence, Tuple, Union

import h5py
import numpy as np
import torch
from torch.utils.data import Dataset

from pearl.common import NodeValueType, same_device


class VariableData(NamedTuple):
    """

    VariableData stores the meta-data of variables as well as the values
    they take in a dataset.

        - value_type: an instance of NodeValueType used for specifying
        whether the variable is CATEGORICAL or CONTINUOUS

        - value: multi-dimensional tensor containing the values the
        variable takes in the dataset.  The tensor contains one
        dimension for each plate in which it occurs.  The left-most
        dimension of the tensor corresponds to the instance plate and
        the remaining dimensions correspond to the other plates in the
        model.

        - domain: A sequence of outcome names. This is used only for
        CATEGORICAL variables and should be None for CONTINUOUS
        variables.
    """

    value_type: NodeValueType
    value: torch.Tensor
    discrete_domain: Optional[Sequence[str]] = None


class BayesianNetworkDataset(Dataset):
    """
    A dataset suitable for use by the BayesianNetwork.  We assume that
    the variables have atleast one batch dimension (the one
    corresponding to the instances).  There can be more based on the
    plates in the graphical model.
    """

    def __init__(
        self,
        variable_dict: Mapping[str, VariableData],
    ) -> None:
        """
        :param variable_dict: a dictionary mapping variables to their metadata and values
        """

        num_instances, device = self._validate_variable_dict(variable_dict)
        self.variable_dict = dict(variable_dict)
        self.num_instances = num_instances
        self.device = device

    def _validate_variable_dict(self, variable_dict):
        for varname, variable in variable_dict.items():
            if not variable.value.ndim > 0:
                raise ValueError(f"{varname} should have atleast one plate dimension")
            if variable.value_type == NodeValueType.CATEGORICAL:
                if variable.discrete_domain is None:
                    raise ValueError(f"{varname} is categorical and should have domain")
            if variable.value_type == NodeValueType.CONTINUOUS:
                if variable.discrete_domain is not None:
                    raise ValueError(
                        f"{varname} is continuous and should not have domain"
                    )
            if variable.value.dtype != torch.float:
                raise ValueError(f"{varname} value tensor should be of type float")

        all_num_instances = [
            variable.value.size(dim=0) for variable in variable_dict.values()
        ]
        if len(set(all_num_instances)) > 1:
            raise ValueError(
                f"all variables should have equal number of instances - {all_num_instances}"
            )
        all_devices = [variable.value.device for variable in variable_dict.values()]
        if not all(same_device(d, all_devices[0]) for d in all_devices):
            raise ValueError(
                f"all variables should use same device for value tensors - {all_devices}"
            )
        return all_num_instances[0], all_devices[0]

    def __len__(self) -> int:
        """
        Return the number of instances in the dataset.
        """
        return self.num_instances

    def subseq(self, i: int, j: int = None) -> "BayesianNetworkDataset":
        """
        Return the instances in the range (i, j) as a BayesianNetworkDataset object.
        If 'j' is omitted we return the instances starting from i.
        """
        variable_dict = dict()
        for v in self.variable_dict:
            variable_dict[v] = VariableData(
                self.variable_dict[v].value_type,
                self.variable_dict[v].value[i:j].clone().detach(),
                self.variable_dict[v].discrete_domain,
            )
        return BayesianNetworkDataset(variable_dict)

    def __getitem__(self, key) -> torch.Tensor:
        """
        Get the value of a variable in the BayesianNetworkDataset.
        The key is of the form (var_name, *idx) where var_name is the
        name of the variable and idx is an optional sequence of
        indices into the value tensor.
        """
        if isinstance(key, str) and key in self.variable_dict:
            return self.variable_dict[key].value
        elif isinstance(key, tuple) and key[0] in self.variable_dict:
            return self.variable_dict[key[0]].value[key[1:]]
        else:
            raise KeyError(f"unrecognized key {key}")

    def __setitem__(self, key, values) -> None:
        """
        Set the value of a variable in the BayesianNetworkDataset.
        The key is of the form (var_name, *idx) where var_name is the
        name of the variable and idx is an optional sequence of
        indices into the value tensor.
        """
        if isinstance(key, str) and key in self.variable_dict:
            self.variable_dict[key].value[:] = values
        elif isinstance(key, tuple) and key[0] in self.variable_dict:
            self.variable_dict[key[0]].value[key[1:]] = values
        else:
            raise KeyError(f"unrecognized key {key}")

    def select(self, assignment) -> torch.Tensor:
        """
        Compute mask that selects rows which are consistent with the
        assignment.  This operation is only supported for datasets
        which have no plate dimension other than the one corresponding
        to instances.
        """
        assert all(variable.value.ndim == 1 for variable in self.variable_dict.values())
        mask = torch.ones(self.num_instances, device=self.device, dtype=torch.bool)
        for varname, value in assignment.items():
            assert varname in self.variable_dict
            if isinstance(value, str):
                assert (
                    self.variable_dict[varname].value_type == NodeValueType.CATEGORICAL
                )
                assert value in self.variable_dict[varname].discrete_domain
                mask = mask & torch.eq(
                    self.variable_dict[varname].value,
                    self.variable_dict[varname].discrete_domain.index(value),
                )
            elif isinstance(value, Iterable):
                # This is a hacky way to support selecting with torch methods,
                # e.g. {"continuous_var": (torch.gt, 0.0)}
                mask = mask & value[0](self.variable_dict[varname].value, *value[1:])
            else:  # pragma: no cover
                raise ValueError(f"Can't handle select value {value}")
        return mask

    def discrete_domain_size(self, varname) -> int:
        """
        Return the size of the domain of a discrete variable
        """
        return len(self.discrete_domain(varname))

    def discrete_domain(self, varname) -> Sequence[str]:
        """
        Return the domain of a discrete variable
        """
        assert varname in self.variable_dict
        assert self.variable_dict[varname].value_type == NodeValueType.CATEGORICAL
        return self.variable_dict[varname].discrete_domain

    def split(
        self, partitions: Union[Tuple[int, ...], Tuple[np.ndarray, ...]]
    ) -> Sequence["BayesianNetworkDataset"]:
        """
        Create a split of the dataset based on a tuple of sizes or
        indices.  If a tuple of sizes is given a random split is
        returned.  On the other hand if a tuple of index sets is
        given, the split uses exactly the specified index sets.
        """
        if isinstance(partitions[0], int):
            assert sum(partitions) == len(self)
            assert all([size > 0 for size in partitions])
            indices = torch.randperm(sum(partitions)).tolist()
            index_sets = []
            for offset, length in zip(accumulate(partitions), partitions):
                index_sets.append(indices[offset - length : offset])
        elif isinstance(partitions[0], np.ndarray):
            assert all([p.ndim == 1 for p in partitions])
            assert reduce(np.intersect1d, partitions).size == 0
            assert np.array_equal(
                reduce(np.union1d, partitions),
                np.arange(len(self)),
            )
            index_sets = partitions
        else:
            raise ValueError(
                "partitions should be tuple of ints of tuple of numpy arrays"
            )

        subsets = []
        for index_set in index_sets:
            subset_variables_dict = dict()
            for varname, variable in self.variable_dict.items():
                subset_variables_dict[varname] = VariableData(
                    variable.value_type,
                    variable.value[index_set],
                    variable.discrete_domain,
                )
            subsets.append(BayesianNetworkDataset(subset_variables_dict))
        return subsets

    def project(self, variables: List[str]) -> "BayesianNetworkDataset":
        """
        Project the dataset onto the specified variables.  The
        resulting dataset contains the union of the plates of these
        variables.
        """
        assert len(variables) > 0
        assert all([v in self.variable_dict for v in variables])
        projected_variable_dict = {v: self.variable_dict[v] for v in variables}
        return BayesianNetworkDataset(
            variable_dict=projected_variable_dict,
        )

    def copy(self) -> "BayesianNetworkDataset":
        """Return a shallow copy"""
        return copy.copy(self)

    def deepcopy(self) -> "BayesianNetworkDataset":
        """Return a deep copy"""
        return copy.deepcopy(self)

    def is_categorical(self, var: str) -> bool:
        """
        Return True if the specified variable is a categorical
        variable in the dataset.
        """
        return self.variable_dict[var].value_type == NodeValueType.CATEGORICAL

    def to_hdf5(self, fname: str) -> None:
        """
        Save to hdf5 file.  This method converts all tensors to numpy
        arrays and stores them in the hdf5 file together with
        metadata.  Device information and id tensors are not retained
        when saving to hdf5.
        """
        with h5py.File(fname, "w") as hdf:
            for k, v in self.variable_dict.items():
                hdf5_dataset = hdf.create_dataset(k, data=v.value.cpu().numpy())
                hdf5_dataset.attrs["type"] = v.value_type
                if v.value_type == NodeValueType.CATEGORICAL:
                    hdf5_dataset.attrs["domain"] = v.discrete_domain

    @classmethod
    def from_hdf5(cls, fname: str) -> "BayesianNetworkDataset":
        """
        Instantiate a BayesianNetworkDataset object from an hdf5 file.
        Since device information and id tensors aren't stored when a
        BayesianNetworkDataset object is saved to hdf5 file, these get
        default values.  The user can use .cuda() method if the
        tensors need to moved to a cuda device.
        """
        variable_dict = dict()
        with h5py.File(fname, "r") as hdf:
            for k in hdf.keys():
                hdf5_dataset = hdf.get(k)
                if hdf5_dataset.attrs["type"] == NodeValueType.CATEGORICAL:
                    variable_dict[k] = VariableData(
                        NodeValueType.CATEGORICAL,
                        torch.tensor(hdf5_dataset).float(),
                        hdf5_dataset.attrs["domain"].tolist(),
                    )
                else:
                    variable_dict[k] = VariableData(
                        NodeValueType.CONTINUOUS,
                        torch.tensor(hdf5_dataset).float(),
                        None,
                    )
        return cls(variable_dict)

    def __eq__(self, other: "BayesianNetworkDataset") -> bool:
        """
        Test equality of two BayesianNetworkDataset objects.
        """
        if not same_device(self.device, other.device):
            return False
        if set(self.variable_dict.keys()) != set(other.variable_dict.keys()):
            return False
        for k, v in self.variable_dict.items():
            if v.value_type != other.variable_dict[k].value_type:
                return False
            if v.discrete_domain != other.variable_dict[k].discrete_domain:
                return False
            if not torch.equal(self[k], other[k]):
                return False
        return True

    def to(self, device: torch.device) -> "BayesianNetworkDataset":
        """
        Return a copy of the ``BayesianNetworkDataset`` object with
        tensors copied to the specified device.  If the object is
        already using the specified device a shallow copy is returned.

        :param device: torch device to be used for the tensors of the
            copied ``BayesianNetworkDataset`` object.
        """
        if same_device(self.device, device):
            return self.copy()
        else:
            variable_dict = dict()
            for k, v in self.variable_dict.items():
                value_copy = v.value.to(device)
                variable_dict[k] = VariableData(
                    v.value_type,
                    value_copy,
                    v.discrete_domain,
                )
            return BayesianNetworkDataset(
                variable_dict,
            )

    def cpu(self) -> "BayesianNetworkDataset":
        """
        Return a copy of the ``BayesianNetworkDataset`` object with
        tensors copied to CPU memory.  If the object is already using
        CPU memory a shallow copy is returned.
        """
        return self.to(torch.device("cpu", 0))

    def cuda(self, device: torch.device) -> "BayesianNetworkDataset":
        """
        Return a copy of the ``BayesianNetworkDataset`` object with
        tensors copied to the specified CUDA device.  If the object is
        already using the specified device a shallow copy is returned.

        :param device: torch cuda device to be used to store the
            tensors of the copied ``BayesianNetworkDataset`` object.
        """

        if not device.type == "cuda":
            raise ValueError(f"did not receive cuda device as input: {device}")
        return self.to(device)



from typing import Tuple, Union

import torch

from pearl.common import _all_index_combinations_tuple


class TensorIndexMapping:
    """A mapping of indices to other indices (of a tensor).

    A many-to-one mapping of tensor indices serves the purpose of
    capturing context specific independence.  For example if A, B are
    parents of C.  Then there are four different distributions of C
    conditioned on the values of A, B (if all are binary variables).
    But if the value of B irrelevant when A is 1, then {A=1, B=0} and
    {A=1, B=1} can both be mapped to {A=1, B=0}.  In terms of tensor
    indices the indices [1,0] and [1,1] are both mapped to [1,0].
    """

    # TODO: rename size to something more intuitive
    def __init__(self, size: torch.Size, device: torch.device = None) -> None:
        """Initially we map each index to itself.

        :param size: size of the tensor for which we are maintaining
            the mapping
        :param device: device on which the mapping should be stored
        """

        assert len(size) > 0
        assert all([s > 1 for s in size])
        if device is None:
            device = torch.device("cpu")
        index_ranges = map(lambda s: torch.arange(s, device=device), size)
        self.mapping = torch.cartesian_prod(*index_ranges).reshape(*size, len(size))
        self.size = size

    def add_mapping(
        self, input_indices: Tuple[Union[int, slice]], output_index: Tuple[int]
    ) -> None:
        """Map input indices to the output index.

        :param input_indices: the range of indices to be mapped
        :param output_index: a single canonical index
        """

        assert len(input_indices) == len(output_index) == len(self.size)
        self.mapping[input_indices] = torch.tensor(
            output_index, device=self.mapping.device
        )

    def map(self, cpd: torch.Tensor) -> torch.Tensor:
        """
        Transform a conditional probability distribution tensor
        according to the mapping.

        :param cpd: tensor whose number/size of batch dimensions is
            suitable for indexing using self.mapping

        :return: transformed cpd
        """
        assert all(cpd.size(dim=i) == self.size[i] for i in range(len(self.size)))
        idx_tuple = _all_index_combinations_tuple(self.size)
        t_idx_tuple = tuple(
            t.view(self.size)
            for t in torch.chunk(self.mapping[idx_tuple], chunks=len(self.size), dim=-1)
        )
        t_cpd = cpd[t_idx_tuple]
        assert t_cpd.shape == cpd.shape
        return t_cpd



import logging
from dataclasses import dataclass
from enum import Enum, IntEnum
from typing import Tuple

import pyro
import torch


@dataclass
class Plate:
    """
    Class to represent plates in graphical models.
    Stores corresponding pyro plate object.
    """

    name: str
    dim: int

    # underlying pyro plate
    pyro_plate: pyro.plate = None


class SamplingMode(Enum):
    MODEL = 0
    GUIDE = 1
    MAP_POSTERIOR = 2


class NodeValueType(IntEnum):
    CATEGORICAL = 0
    CONTINUOUS = 1


class QueryType(Enum):
    CONJUNCTIVE = 0
    DISJUNCTIVE = 1


def same_device(device1, device2) -> bool:  # pragma: no cover
    if device1.type == device2.type == "cpu":
        return True
    elif device1.type == device2.type == "cuda" and (device1.index == device2.index):
        return True
    else:
        return False


def _all_index_combinations(domain_sizes: torch.Size) -> torch.Tensor:

    """Enumerate the possible combinations of indices for "domain_sizes". e.g. if
    the node has 2 parents with 2 and 3 possible values respectively, returns
    [[0, 0], [0, 1], [0, 2], [1, 0], [1, 1], [1, 2]]

    reshape() is required to ensure the return value has the correct dims in
    the case where there is a single domain size.
    """
    return torch.cartesian_prod(*(torch.arange(s) for s in domain_sizes)).reshape(
        torch.prod(torch.tensor(domain_sizes)), len(domain_sizes)
    )


def _all_index_combinations_tuple(domain_sizes: torch.Size) -> Tuple[torch.Tensor]:
    index_combinations = _all_index_combinations(domain_sizes)
    return tuple(
        t.squeeze()
        for t in torch.chunk(index_combinations, chunks=len(domain_sizes), dim=-1)
    )


def tensor_accuracy(tensor1: torch.Tensor, tensor2: torch.Tensor) -> float:
    if not tensor1.shape == tensor2.shape:
        raise ValueError(
            f"Shapes of input tensors are not equal: {tensor1.shape} vs {tensor2.shape}"
        )

    return float(
        torch.sum(tensor1 == tensor2) / torch.prod(torch.tensor(tensor1.shape)).float()
    )


def get_logger(name, level=None):
    logger = logging.getLogger(name)
    level = level or logging.INFO
    logger.setLevel(level)

    ch = logging.StreamHandler()
    ch.setLevel(level)
    logger.addHandler(ch)

    return logger



import math
from functools import reduce
from itertools import combinations, product
from typing import AbstractSet, Mapping

import networkx as nx
import torch
from tqdm import tqdm

from pearl.bayesnet import BayesianNetwork
from pearl.data import BayesianNetworkDataset
from pearl.nodes.categorical import CategoricalNodeWithDirichletPrior


def categorical_tree_augmented_naive_bayes(
    name: str,
    dataset: BayesianNetworkDataset,
    class_variable: str,
    augmenting_tree_root: str,
    features: AbstractSet[str] = None,
) -> BayesianNetwork:
    """
    Create a categorical tree augmented Naive Bayes model using
    information from the dataset.  This function merely instantiates a
    BayesianNetwork object and doesn't train it on the dataset.

    :param name: name of the BayesianNetwork object to be created
    :param dataset: a BayesianNetworkDataset object that provides
        information for creating the tree augmented naive bayes model.
    :param class_variable: class variable of the model.
    :param augmenting_tree_root: feature to be used as the root of the
        augmenting tree.  Should be included in 'features' parameter
        if it is not None.
    :param features: Set of features to be included in the model.  If
        this argument is None, then all the categorical features in
        the dataset are used.
    """

    if features is None:
        features = set(
            f for f in dataset.variable_dict if dataset.is_categorical(f)
        ) - {class_variable}

    if augmenting_tree_root not in features:
        raise ValueError(
            f"{augmenting_tree_root} should be part of the feature variables"
        )

    for var in features | {class_variable, augmenting_tree_root}:
        if not dataset.is_categorical(var):
            raise ValueError(f"{var} should be a categorical variable in the dataset")

    projected_dataset = dataset.project(
        list(features | {class_variable, augmenting_tree_root})
    )

    model = BayesianNetwork(name, dataset.device)
    model.add_variable(
        node_class=CategoricalNodeWithDirichletPrior,
        node_name=class_variable,
        node_parents=[],
        plates=[],
        domain=dataset.discrete_domain(class_variable),
    )
    model.add_variable(
        node_class=CategoricalNodeWithDirichletPrior,
        node_name=augmenting_tree_root,
        node_parents=[class_variable],
        plates=[],
        domain=dataset.discrete_domain(augmenting_tree_root),
    )

    for pred, succ in augmenting_tree(
        projected_dataset, class_variable, augmenting_tree_root
    ).edges:
        model.add_variable(
            node_class=CategoricalNodeWithDirichletPrior,
            node_name=succ,
            node_parents=[class_variable, pred],
            plates=[],
            domain=dataset.discrete_domain(succ),
        )
    return model


def augmenting_tree(
    dataset: BayesianNetworkDataset, class_variable: str, augmenting_tree_root: str
) -> nx.Graph:
    """
    Construct the directed tree whose edges are used for augmenting
    the Naive Bayes model.  Its constructed as follows

        - create a complete graph whose edges are weighted by
          conditional mutual information

        - construct the maximum spanning tree of the above graph

        - make the maximum spanning tree directed by using
          'augmenting_tree_root' as the root of BFS

    :param dataset: BayesianNetworkDataset object used for
        constructing maximum spanning tree.
    :param class_variable: class variable
    :param augmenting_tree_root: variable to be used as the root of
        the augmenting tree
    """

    g = nx.Graph()
    features = set(dataset.variable_dict.keys()) - {class_variable}
    num_total_instantiations = reduce(
        lambda x, y: x * y,
        [dataset.discrete_domain_size(v) for v in dataset.variable_dict],
    )
    for node1, node2 in tqdm(combinations(features, 2)):
        cmi = conditional_mutual_information(
            dataset, node1, node2, class_variable, num_total_instantiations
        )
        g.add_edge(node1, node2, weight=cmi)
    max_spanning_tree = nx.maximum_spanning_tree(g)
    return nx.bfs_tree(max_spanning_tree, augmenting_tree_root)


def conditional_mutual_information(
    dataset: BayesianNetworkDataset,
    x: str,
    y: str,
    z: str,
    num_total_instantiations: int,
) -> float:
    """
    Compute conditional mutual information using the dataset.

    :param dataset: dataset used for computing conditional mutual information
    :param x: random variable (see below formula)
    :param y: random variable (see below formula)
    :param z: random variable (see below formula)
    :param num_total_instantiations: total number of possible
        instantiations of random variables in the dataset

    I_p(X, Y | Z) = sum_{x, y, z} P(x, y, z) * log(P(x, y | z) / P(x | z) * P(y | z))
    """

    x_domain = dataset.discrete_domain(x)
    y_domain = dataset.discrete_domain(y)
    z_domain = dataset.discrete_domain(z)

    cmi = 0.0  # conditional mutual information
    for xval, yval, zval in product(x_domain, y_domain, z_domain):
        term1 = empirical_probability(
            dataset, {x: xval, y: yval, z: zval}, num_total_instantiations
        )
        term2 = log_empirical_conditional_probability(
            dataset, {x: xval, y: yval}, {z: zval}, num_total_instantiations
        )
        term3 = log_empirical_conditional_probability(
            dataset, {x: xval}, {z: zval}, num_total_instantiations
        )
        term4 = log_empirical_conditional_probability(
            dataset, {y: yval}, {z: zval}, num_total_instantiations
        )
        cmi += term1 * (term2 - term3 - term4)
    return cmi


def log_empirical_conditional_probability(
    dataset: BayesianNetworkDataset,
    assignment1: Mapping[str, str],
    assignment2: Mapping[str, str],
    num_total_instantiations: int,
) -> float:
    """
    Compute log P(assignment1 | assignment2)

    :param dataset: dataset used for computing conditional probability
    :param assignment1: variable assignment
    :param assignment2: variable assignment
    :param num_total_instantiations: total number of possible
        instantiations of random variables in the dataset
    """
    term1 = empirical_probability(
        dataset, {**assignment1, **assignment2}, num_total_instantiations
    )
    term2 = empirical_probability(dataset, assignment2, num_total_instantiations)
    return math.log2(term1) - math.log2(term2)


def empirical_probability(
    dataset: BayesianNetworkDataset,
    assignment: Mapping[str, str],
    num_total_instantiations: int,
) -> float:
    """
    Compute the empirical probability of a random variable assignment
    according to the dataset.  Additive smoothing is used to avoid
    zero probabilities.

    :param dataset: dataset used for computing empirical probability
    :param assignment: random variable assignment for which empirical
        probability is to be computed
    :param num_total_instantiations: total number of possible
        instantiations of random variables in the dataset
    """
    smoothing_multiplicative_factor = num_total_instantiations
    for var in assignment.keys():
        smoothing_multiplicative_factor /= dataset.discrete_domain_size(var)
    count = (
        int(torch.sum(dataset.select(assignment)))
        + 1 / num_total_instantiations * smoothing_multiplicative_factor
    )
    return count / (len(dataset) + 1)



from typing import Any, Callable, List, Mapping, Sequence

import pyro.distributions as dist
import torch

from pearl.common import NodeValueType, SamplingMode
from pearl.nodes.bayesnetnode import Node


class DeterministicNode(Node):
    """Node whose value is a deterministic function of its parents,
    as defined by the func input parameter.

    func must take in a Dict[str, torch.tensor] with
    keys corresponding to parent names and values to parent values
    """

    def __init__(
        self,
        name: str,
        func: Callable,
        node_value_type: NodeValueType,
        plates: List[str],
        parents: Sequence[Node],
        domain: Sequence[str] = None,
        device: torch.device = torch.device("cpu"),
        observed: bool = False,
    ) -> None:
        super().__init__(
            name,
            plates,
            parents,
            observed=observed,
            device=device,
        )

        if node_value_type == NodeValueType.CATEGORICAL and domain is None:
            raise ValueError(
                f"domain must be set for categorical deterministic node {self.name}"
            )
        if node_value_type == NodeValueType.CONTINUOUS and domain is not None:
            raise ValueError(
                f"domain should not be set for continuous deterministic node {self.name}"
            )

        self.observed = observed
        self.node_value_type = node_value_type
        self.domain = domain
        self.func = func

    @property
    def domain_size(self):
        return len(self.domain) if self.domain else None

    def get_sample_dist(
        self, sampling_mode: SamplingMode, *parent_values: torch.Tensor
    ) -> dist.Distribution:
        """
        Compute the deterministic function based on the parents' values
        and return the corresponding Delta distribution.
        """

        d_parent_values = {
            parent.name: value for parent, value in zip(self.parents, parent_values)
        }

        f_output = self.func(d_parent_values)

        return dist.Delta(f_output)

    def value_type(self) -> NodeValueType:
        return self.node_value_type

    def overwrite_param_store(self) -> None:
        pass

    @classmethod
    def process_yaml_spec(
        cls, bayesnet, yaml_encoding: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        kwargs = super().process_yaml_spec(bayesnet, yaml_encoding)
        if "domain" in yaml_encoding:
            kwargs["domain"] = yaml_encoding["domain"]
        return kwargs


class Exponential(DeterministicNode):
    """
    Node which implements the 'torch.exp' function
    """

    def __init__(
        self,
        name: str,
        plates: List[str],
        parents: Sequence[Node],
        device: torch.device = torch.device("cpu"),
        observed: bool = False,
    ) -> None:
        super().__init__(
            name=name,
            func=torch.exp,
            node_value_type=NodeValueType.CONTINUOUS,
            plates=plates,
            parents=parents,
            device=device,
            observed=observed,
        )

    def to_yaml_encoding(self) -> Mapping[str, Any]:
        yaml_encoding = dict(super().to_yaml_encoding())
        return yaml_encoding


class Sum(DeterministicNode):
    """
    Node which implements the 'torch.sum' function
    """

    def __init__(
        self,
        name: str,
        plates: List[str],
        parents: Sequence[Node],
        device: torch.device = torch.device("cpu"),
        observed: bool = False,
    ) -> None:
        super().__init__(
            name=name,
            func=torch.sum,
            node_value_type=NodeValueType.CONTINUOUS,
            plates=plates,
            parents=parents,
            device=device,
            observed=observed,
        )

    def to_yaml_encoding(self) -> Mapping[str, Any]:
        yaml_encoding = dict(super().to_yaml_encoding())
        return yaml_encoding



from typing import AbstractSet

from pearl.bayesnet import BayesianNetwork
from pearl.data import BayesianNetworkDataset
from pearl.nodes.categorical import CategoricalNodeWithDirichletPrior


def categorical_naive_bayes_model(
    name: str,
    dataset: BayesianNetworkDataset,
    class_variable: str,
    features: AbstractSet[str] = None,
) -> BayesianNetwork:
    """
    Create a categorical Naive Bayes model using information from the
    dataset.  This function merely instantiates a BayesianNetwork
    object and doesn't train it on the dataset.

    :param name: name of the BayesianNetwork object to be
        created
    :param dataset: a BayesianNetworkDataset object that provides
        information about the class variable and features.
    :param class_variable: class variable of the model.
    :param features: Set of features to be included in the model.  If
        this argument is None, then all the categorical features in
        the dataset are used.
    """

    if features is None:
        features = set(
            f for f in dataset.variable_dict if dataset.is_categorical(f)
        ) - {class_variable}

    for var in features | {class_variable}:
        if not dataset.is_categorical(var):
            raise ValueError(f"{var} should be a categorical variable in the dataset")

    model = BayesianNetwork(name=name, device=dataset.device)
    model.add_variable(
        node_class=CategoricalNodeWithDirichletPrior,
        node_name=class_variable,
        node_parents=[],
        plates=[],
        domain=dataset.discrete_domain(class_variable),
    )
    for feature in features:
        model.add_variable(
            node_class=CategoricalNodeWithDirichletPrior,
            node_name=feature,
            node_parents=[class_variable],
            plates=[],
            domain=dataset.discrete_domain(feature),
        )
    return model



import copy
import functools
import itertools
import logging
import time
from pathlib import Path
from typing import Any, Dict, List, Mapping, Optional, Tuple, Type, Union

import networkx as nx
import pyro
import pyro.optim
import torch
import yaml
from pyro.infer import SVI, Trace_ELBO, config_enumerate, infer_discrete

from pearl.common import (
    NodeValueType,
    Plate,
    QueryType,
    SamplingMode,
    get_logger,
    same_device,
)
from pearl.data import BayesianNetworkDataset, VariableData
from pearl.hugin import hugin_potential_string
from pearl.nodes.bayesnetnode import Node
from pearl.nodes.categorical import (
    CategoricalNodeWithDirichletPrior,
    GeneralizedLinearNode,
)
from pearl.nodes.continuous import (
    ConditionalLinearGaussianNode,
    ContinuousNodeWithNormalDistribution,
)
from pearl.nodes.deterministic import Exponential, Sum

DATA_LOOP = "data_loop"
PREDICTIVE_SAMPLES = "predictive_samples"
ENCODING_VERSION = "v1.0"  # TODO: versioning system that updates
# automatically (for e.g. using hash of class attributes?)
NODE_CLASS = {
    "CategoricalNodeWithDirichletPrior": CategoricalNodeWithDirichletPrior,
    "ContinuousNodeWithNormalDistribution": ContinuousNodeWithNormalDistribution,
    "GeneralizedLinearNode": GeneralizedLinearNode,
    "ConditionalLinearGaussianNode": ConditionalLinearGaussianNode,
    "Exponential": Exponential,
    "Sum": Sum,
}


class BayesianNetwork:
    """
    Bayesian Network.

    We make the following assumptions about the plates in the
    graphical model:

        - Plates have unique names and dimensions.  The name
          "data_loop" is reserved for use by the library and is used
          for iid samples generated from the BayesianNetwork.  The
          user is required to specify all the plates in the model
          except the "data_loop".

        - If there are 'N' plates (excluding the "data_loop") then
          they should use dimensions -N, -N+1, ..., -1.

        - During training and prediction, the "data_loop" plate is
          added to the model with dimension -N-1.

        - If a node is nested in 'M' plates (excluding the
          "data_loop"), the BayesianNetworkDataset object should
          provide a tensor of M+1 dimensions for that node (with the
          left-most dimension corresponding to the "data_loop" and
          remaining dimensions corresponding to the remaining plates
          of the node).

    We verify that each node is included in all the plates of its
    parents.  This is a general requirement of on plate models and it
    also serves to exclude a class of models for which parallel
    enumeration cannot be performed.  We rely on Pyro implementation
    to detect remaining models for which parallel enumeration is not
    feasible.
    """

    def __init__(self, name: str, device: torch.device) -> None:
        """
        :param name: name of the Bayesian Network
        :param device: cpu/cuda device to use for storing CPD tensors of the nodes
        """

        self.name = name
        self.device = device
        self.dag = nx.DiGraph(name=self.name)
        self.plate_dict = dict()

    def add_plate(self, name: str, dim: int) -> None:
        """
        Add a plate to the Bayesian Nework
        """
        if name in self.plate_dict:
            raise ValueError(f"plate with name {name} already exists")
        if name == DATA_LOOP:
            raise ValueError(f"name {DATA_LOOP} is reserved")
        if dim in [plate.dim for plate in self.plate_dict.values()]:
            raise ValueError(f"dimension {dim} is used by another plate")
        self.plate_dict[name] = Plate(name, dim)

    def get_node_object(self, name: str) -> Optional[Node]:
        return self.dag.nodes[name]["node_object"]

    def get_node_dict(self) -> Mapping[str, Node]:
        return {
            node_name: self.get_node_object(node_name) for node_name in self.dag.nodes
        }

    def _add_node(self, node: Node, parents: List[Node]) -> None:
        assert (
            node.name not in self.dag.nodes
        ), f"node {node.name} already added to the network"
        assert all(
            p.name in self.dag.nodes for p in parents
        ), f"parents of {node.name} not yet added to the network"
        self.dag.add_node(node.name, node_object=node)
        for p in parents:
            self.dag.add_edge(p.name, node.name)

    def add_variable(
        self,
        node_class: Type[Node],
        node_name: str,
        node_parents: List[str],
        plates: List[str],
        **kwargs: dict,
    ) -> None:
        """
        Add a random variable to the Bayesian Network.

        :param node_class: Class from which the node should be instantiated
        :param node_name: name of the random variable
        :param node_parents: names of the parents of the random variable
        :param plates: plates in which the node is embedded
        :param kwargs: arguments to be used in the construction of the node
        """

        if not all(p in self.plate_dict for p in plates):
            raise ValueError(f"plates of {node_name} not yet added to the network")

        parents = [self.get_node_object(p) for p in node_parents]
        node = node_class(
            node_name, plates=plates, parents=parents, device=self.device, **kwargs
        )
        self._add_node(node, parents)

    def sample(
        self,
        *,
        sampling_mode: SamplingMode,
        plate_sizes: Mapping[str, int],
        subsample_sizes: Mapping[str, int],
    ) -> None:
        """
        Sample the nodes of the Bayesian Network.  Consider using one
        of these methods instead.  The sampling_mode parameter is
        frozen for these methods and therefore only the remaining
        arguments may be specified.

            - model: sample from the model distribution

            - guide: sample from the guide distribution

            - MAP_posterior: sample from the MAP of the posterior
              guide distribution

        :param sampling_mode: a SamplingMode enum specifying the sampling mode.
        :param plate_sizes: a dictionary that specifies the size of each plate
        :param subsample_sizes: a dictionary that specifies the subsample sizes of plates
        """

        if not all([plate_name in plate_sizes for plate_name in subsample_sizes]):
            raise ValueError(
                "Plates specified in input subsample_sizes must also appear in input plate_sizes"
            )

        if not all(subsample_sizes[k] <= plate_sizes[k] for k in subsample_sizes):
            raise ValueError("subsample size is greater than plate size")

        # create necessary pyro.plate objects
        for plate_name, pearl_plate in self.plate_dict.items():
            pearl_plate.pyro_plate = pyro.plate(
                plate_name,
                size=plate_sizes[plate_name],
                subsample_size=subsample_sizes.get(plate_name, plate_sizes[plate_name]),
                dim=self.plate_dict[plate_name].dim,
            )

        for node_name in nx.topological_sort(self.dag):
            node_object = self.get_node_object(node_name)
            relevant_plates = [self.plate_dict[p] for p in node_object.plates]
            relevant_plates.sort(key=lambda p: p.dim)  # sort the plates by dimension
            parent_values = tuple(
                self.get_node_object(p).value for p in self.dag.predecessors(node_name)
            )
            node_object.sample(relevant_plates, sampling_mode, *parent_values)

        # clear pyro.plate objects
        for plate_name, pearl_plate in self.plate_dict.items():
            pearl_plate.pyro_plate = None

    # Use functools.partialmethod to freeze the sampling_mode argument
    # of sample method. Since functools.partialmethod appends
    # positional arguments to frozen arguments, we require the use of
    # keyword only arguments for sample and the following methods.
    model = functools.partialmethod(sample, sampling_mode=SamplingMode.MODEL)
    guide = functools.partialmethod(sample, sampling_mode=SamplingMode.GUIDE)

    def MAP_posterior(
        self, plate_sizes: Mapping[str, int], subsample_sizes: Mapping[str, int]
    ) -> None:
        """
        Sample the nodes using the MAP of the posterior guide distribution.
        """
        infer_discrete(
            config_enumerate(self.sample),
            first_available_dim=min(
                (p.dim for p in self.plate_dict.values()), default=0
            )
            - 1,
        )(
            sampling_mode=SamplingMode.MAP_POSTERIOR,
            plate_sizes=plate_sizes,
            subsample_sizes=subsample_sizes,
        )

    def train(
        self,
        dataset: BayesianNetworkDataset,
        optimizer: pyro.optim.PyroOptim,
        num_steps: int,
        subsample_size: int = None,
        logdir: str = None,
    ) -> List[float]:

        """Learn the Conditional Probability Distributions of the
        nodes of the Bayesian Network from the dataset.

        :param dataset: dataset for learning CPDs
        :param optimizer: optimizer object to be used with SVI
        :param num_steps: number of steps of SVI
        :param subsample_size: size of minibatch to be sampled from the dataset
        :param logdir: directory where model checkpoints should be saved
        """
        self.validate(dataset)
        pyro.clear_param_store()

        # add the "data_loop" as the outermost plate
        self._add_outer_plate(DATA_LOOP)
        self.set_observed_values(dataset)

        if subsample_size is None:
            subsample_sizes = {DATA_LOOP: len(dataset)}
        else:
            subsample_sizes = {DATA_LOOP: subsample_size}

        plate_sizes = self._infer_plate_sizes(dataset)

        # Bayesian parameter estimation
        svi = SVI(self.model, self.guide, optim=optimizer, loss=Trace_ELBO())
        losses = []
        logger = get_logger(__name__, logging.INFO)
        if logdir is not None and Path(logdir).exists():
            logdir_path = Path(logdir)
        else:
            if logdir is not None:
                logger.warning(
                    f"directory {logdir} doesn't exist, checkpoints will not saved"
                )
            logdir_path = None
        start_time = time.time()
        for i in range(num_steps):
            loss = svi.step(
                plate_sizes=plate_sizes,
                subsample_sizes=subsample_sizes,
            )
            losses.append(loss)
            if (i + 1) % max(1, (num_steps // 10)) == 0:
                # checkpoint model after every 1/10th of num_steps
                logger.info(f"svi_step: {(i + 1)}")
                logger.info(f"elbo: {loss}")
                if logdir_path is not None:
                    checkpoint_file = str(
                        logdir_path.joinpath(f"{self.name}_{(i + 1)}.yaml")
                    )
                    self._remove_plate(DATA_LOOP)
                    to_yaml(self, checkpoint_file)
                    self._add_outer_plate(DATA_LOOP)
        end_time = time.time()
        logger.info(f"training time: {end_time - start_time}")

        # remove the "data_loop" plate
        self._remove_plate(DATA_LOOP)

        return losses

    def predict(
        self,
        dataset: BayesianNetworkDataset,
        target_variables: List[str],
        num_samples: int = 500,
    ) -> Tuple[dict, dict, dict]:
        """Generate predictions using learned conditional probability
        distributions.  It is required that the ``dataset`` have
        appropriately shaped tensors (with random values) for the
        ``target_variables``.  This is used for inferring plate sizes.

        :param dataset: test dataset for which we need predictions
        :param target_variables: list of variables for which we seek
            predictions
        :param num_samples: number of samples to generate from the
            posterior predictive distribution

        :return: A tuple of three dictionaries.  The dictionaries map
                 from target variables to samples, MAP assignments and
                 distributions
        """
        self.validate(dataset)

        # this might be necessary when the object is loaded from a pickle file
        self.overwrite_param_store()

        # add data_loop plate and infer plate sizes from dataset
        self._add_outer_plate(DATA_LOOP)
        plate_sizes = self._infer_plate_sizes(dataset)

        # add predictive_samples plate
        self._add_outer_plate(PREDICTIVE_SAMPLES)
        plate_sizes[PREDICTIVE_SAMPLES] = num_samples

        # temporarily mark the target variables as unobserved
        old_observed_state = dict()
        for target_variable in target_variables:
            node_object = self.get_node_object(target_variable)
            old_observed_state[target_variable] = node_object.observed
            node_object.observed = False

        # set observed values while expanding dimensions to account for the temporary "predictive_samples" plate
        self.set_observed_values(dataset, dims_to_prepend=(num_samples,))

        self.MAP_posterior(
            plate_sizes=plate_sizes,
            subsample_sizes={},
        )

        # extract predictive samples and determine MAP assignment
        samples = dict()
        target_var_shape = {}
        MAP_assignment, assignment_distribution = dict(), dict()
        for target_var in target_variables:
            sample = self.get_node_object(target_var).value
            plates_dim_size = [
                (self.plate_dict[plate].dim, sample.shape[self.plate_dict[plate].dim])
                for plate in self.get_node_object(target_var).plates
            ]
            target_var_shape[target_var] = tuple(
                [sz for (dim, sz) in sorted(plates_dim_size)]
            )
            samples[target_var] = (
                sample.clone().detach().view(target_var_shape[target_var])
            )

            node_object = self.get_node_object(target_var)
            if node_object.value_type() == NodeValueType.CATEGORICAL:
                counts = torch.zeros(
                    target_var_shape[target_var][
                        1:
                    ]  # need to skip PREDICTIVE_SAMPLES plate
                    + (node_object.domain_size,),
                    device=self.device,
                )
                for i in range(node_object.domain_size):
                    counts[..., i] = torch.eq(samples[target_var], i).sum(dim=0)
                MAP_assignment[target_var] = torch.argmax(counts, dim=-1)
                assignment_distribution[target_var] = counts / num_samples
            else:
                MAP_assignment[target_var] = samples[target_var].mean(dim=0)
                assignment_distribution[target_var] = samples[target_var]

        # remove data_loop and predictive_samples plates
        self._remove_plate(DATA_LOOP)
        self._remove_plate(PREDICTIVE_SAMPLES)

        # undo changes to target variables
        for target_variable in target_variables:
            node_object = self.get_node_object(target_variable)
            node_object.observed = old_observed_state[target_variable]

        return (samples, MAP_assignment, assignment_distribution)

    def conditional_query(
        self, query: str, evidence: str, query_type: QueryType, num_samples: int
    ) -> Dict[str, Union[float, torch.Tensor]]:
        """
        Compute the conditional probability P(query|evidence).

        The ``evidence`` is an assignment of values to some random
        variables.

        The ``query`` can either be a joint assignment to a set of
        random variables (i.e., conjunctive) or be a disjunction of
        variable assignments.  The distinction is made by
        ``query_type`` argument.

        :param query: string containing probabilistic query
        :param evidence: string containing evidence for the
            probabilistic query
        :param query_type: type of the query (conjunctive vs
            disjunctive)
        :param num_samples: number of samples to be used for
            estimating the conditional probability

        :return: A dictionary whose keys are variable assignments to
                 discrete variables in the query.  The values are
                 either estimated conditional probabilities or
                 tensors.  When the query does not contain continuous
                 random variables, the values are conditional
                 probabilities of the variable assignments to the
                 query variables.  When the query contains continuous
                 random variables, the values are tensors containing
                 the sampled values of continuous random variables
                 from the conditional distribution.
        """

        # check that there are no plates
        if self.plate_dict:
            raise Exception(
                f"Conditional queries for plated graphical models are not supported: {self.plate_dict}"
            )

        # parse evidence
        evidence_dict = self.parse_evidence(evidence)

        # parse query
        (
            q_constraints,
            q_unconstrained_discrete_vars,
            q_unconstrained_continuous_vars,
        ) = self.parse_query(query)
        q_constraints_vars = [var for (var, _) in q_constraints]
        q_constraints_vals = [val for (_, val) in q_constraints]

        # validate query
        if query_type == QueryType.DISJUNCTIVE:
            if q_unconstrained_discrete_vars or q_unconstrained_continuous_vars:
                raise ValueError(
                    f"conditional disjunctive query with unconstrained variables: {query}"
                )
        else:
            # each discrete variable should have a single constraint
            if len(set(q_constraints_vars)) != len(q_constraints_vars):
                raise ValueError(f"some variables have multiple constraints in {query}")

        # constrained and unconstrained variables should be disjoint
        if set(q_constraints_vars) & set(q_unconstrained_discrete_vars):
            raise ValueError(
                f"some variables are both constrained and unconstrained in {query}"
            )
        # unconstrained variables should not be repeated
        if (
            len(set(q_unconstrained_discrete_vars))
            != len(q_unconstrained_discrete_vars)
        ) or (
            len(set(q_unconstrained_continuous_vars))
            != len(q_unconstrained_continuous_vars)
        ):
            raise ValueError(f"some unconstrained variables are repeated in {query}")

        # collect the domains of unconstrained discrete variables
        q_unconstrained_discrete_var_domains = []
        for var in q_unconstrained_discrete_vars:
            node_object = self.get_node_object(var)
            q_unconstrained_discrete_var_domains.append(range(len(node_object.domain)))

        # ensure guide parameters are set
        for node_object in self.get_node_dict().values():
            node_object.sample_guide_cpd()

        # create BayesianNetworkDataset containing a single row using the evidence
        variable_dict = dict()
        for k, node_object in self.get_node_dict().items():
            if k in evidence_dict:
                dataset_value = torch.tensor(
                    [evidence_dict[k]], device=self.device
                ).float()
            else:
                dataset_value = torch.tensor([0.0], device=self.device)
            if node_object.value_type() == NodeValueType.CATEGORICAL:
                variable_dict[k] = VariableData(
                    node_object.value_type(), dataset_value, node_object.domain
                )
            else:
                variable_dict[k] = VariableData(
                    node_object.value_type(), dataset_value, None
                )
        dataset = BayesianNetworkDataset(variable_dict)

        # temporarily mark evidence variables as observed and the rest as unobserved
        old_observed_state = dict()
        for k, node_object in self.get_node_dict().items():
            old_observed_state[k] = node_object.observed
            if k not in evidence_dict:
                node_object.observed = False
            else:
                node_object.observed = True

        # generate predictive samples
        samples, _, _ = self.predict(
            dataset=dataset,
            target_variables=q_constraints_vars
            + q_unconstrained_discrete_vars
            + q_unconstrained_continuous_vars,
            num_samples=num_samples,
        )

        # undo changes to the variables
        for k, node_object in self.get_node_dict().items():
            node_object.observed = old_observed_state[k]

        # process the samples to compute query probabilities
        result_dict = dict()
        if q_unconstrained_discrete_vars:
            for vals in itertools.product(*q_unconstrained_discrete_var_domains):
                entry = tuple(q_constraints_vals) + vals
                result_dict[entry] = self._conditional_query_helper(
                    query_type,
                    q_constraints_vars + q_unconstrained_discrete_vars,
                    q_constraints_vals + list(vals),
                    q_unconstrained_continuous_vars,
                    samples,
                )
        else:
            entry = tuple(q_constraints_vals)
            result_dict[entry] = self._conditional_query_helper(
                query_type,
                q_constraints_vars,
                q_constraints_vals,
                q_unconstrained_continuous_vars,
                samples,
            )
        return result_dict

    def conditional_conjunctive_query(
        self, query: str, evidence: str, num_samples: int
    ) -> Dict[str, Union[float, torch.Tensor]]:
        """
        Compute P(query|evidence) where ``query`` represents a
        conjunction of events.  This method is supported only by
        BayesianNetworks without plates.

        :param query: String containing the conjunctive query.  It has
            the form of a comma separated list of conjuncts.  Each
            conjunct is either an un-constrained random variable, or
            an equality constrained categorical random variable.

            Examples:

                - "EV0016 = True"

                - "EV0016 = True, EV0014, Dma.Score"

            Spaces are allowed around the comma separator and the
            equality operator.

        :param evidence: a string containing the evidence on the
            random variables.  It has the form of a comma separated
            list of equality constraints on random variables.

            Examples:

                - "Interpretation = pathogenic"

                - "Interpretation = pathogenic, Dma.Score = 0.7"

        :return: A dictionary that maps from categorical query
                 variable assignments to probabilities or tensors.
                 Values are probabilities when the query does not
                 contain continuous random variables.  Values are
                 tensors of sampled values of the continuous random
                 variables, when they are present in the query.
        """
        return self.conditional_query(
            query, evidence, QueryType.CONJUNCTIVE, num_samples
        )

    def conditional_disjunctive_query(
        self, query: str, evidence: str, num_samples: int
    ) -> float:
        """
        Compute P(query|evidence) where ``query`` represents a
        disjunction of events.  This method is supported only by
        BayesianNetworks without plates.

        :param query: String containing the conjunctive query.  It has
            the form of a comma separated list of disjuncts.  Each
            disjunct is an equality constrained categorical random
            variable.

            Examples:

                - "EV0016 = True"

                - "EV0016 = True, EV0014=False"

            Spaces are allowed around the comma separator and the
            equality operator.

        :param evidence: a string containing the evidence on the
            random variables.  It has the form of a comma separated
            list of equality constraints on random variables.

            Examples:

                - "Interpretation = pathogenic"

                - "Interpretation = pathogenic, Dma.Score = 0.7"

        :return: The conditional probability of the query.
        """
        result_dict = self.conditional_query(
            query, evidence, QueryType.DISJUNCTIVE, num_samples
        )
        if len(result_dict) > 1:
            raise Exception(
                f"result dictionary of disjunctive query has more than one entry: {result_dict}"
            )
        return float(next(iter(result_dict.values())))

    def _conditional_query_helper(
        self,
        query_type: QueryType,
        categorical_variables: List[str],
        values: List[int],
        continuous_variables: List[str],
        samples: dict,
    ) -> Union[float, torch.Tensor]:
        """
        Helper function to compute conditional query probabilities
        from samples.  For the given assignment of ``values`` to
        (discrete) ``categorical_variables``, the helper function
        returns either the fraction of samples consistent with the
        assignment or returns a tensor with the consistent samples.
        The choice is made based on whether the query includes
        continuous random variables.  If there are no continuous
        random variables, a fraction is returned as the estimate of
        the conditional probability.  Otherwise the consistent samples
        of the continuous random variables are returned in the tensor.

        :param categorical_variables: list of categorical randomc
            variables in the query
        :param values: list of values of categorical variables
        :param continuous_variables: list of continuous variables in
            the query
        :param samples: samples from the predictive distribution
        """
        num_samples = len(next(iter(samples.values())))
        if query_type == QueryType.CONJUNCTIVE:
            mask = torch.ones(num_samples).bool()
            for k, v in zip(categorical_variables, values):
                mask = mask & torch.eq(samples[k].squeeze(), v)
        else:
            mask = torch.zeros(num_samples).bool()
            for k, v in zip(categorical_variables, values):
                mask = mask | torch.eq(samples[k].squeeze(), v)

        if continuous_variables:
            return torch.stack(
                [samples[k].squeeze()[mask] for k in continuous_variables],
                dim=-1,
            )
        else:
            return float(torch.sum(mask)) / num_samples

    def set_observed_values(
        self, dataset: BayesianNetworkDataset, dims_to_prepend: Tuple[int] = ()
    ) -> None:
        """Annotate the observed nodes in the graph with observed values. If unobserved then value is set to None.

        :param dataset: the BayesianNetworkDataset object used for annotating the graph with observed values
        :param dims_to_prepend: tuple of dimensions to be prepended to value tensors (used by predict())
        """

        for node_name, node_object in self.get_node_dict().items():
            value = None
            if node_object.observed:
                value = dataset[node_name].clone().detach()

                # prepend dimensions if needed, as in prediction
                value = value.expand(dims_to_prepend + value.shape)

            node_object.set_observed_value(value)

    def generate_dataset(self, plate_sizes: Mapping[str, int]):
        """

        :param plate_sizes: a dictionary that specifies the size of each plate
        :return: dataset object sampled from the model distribution
        """

        if DATA_LOOP not in plate_sizes:
            raise ValueError(
                f"Must specify size for plate {DATA_LOOP} when generating a dataset."
            )

        self._add_outer_plate(DATA_LOOP)

        # temporarily mark all variables as unobserved
        old_observed_state = dict()
        for node_name, node_object in self.get_node_dict().items():
            old_observed_state[node_name] = node_object.observed
            node_object.observed = False

        # sample the model with no observations
        self.model(plate_sizes=plate_sizes, subsample_sizes={})

        variable_dict = {}
        for node_name, node_object in self.get_node_dict().items():
            domain = (
                node_object.domain
                if node_object.value_type() == NodeValueType.CATEGORICAL
                else None
            )
            ordered_plates = list(
                zip(*sorted([(self.plate_dict[p].dim, p) for p in node_object.plates]))
            )[1]
            shape = [plate_sizes[p] for p in ordered_plates]
            value = torch.reshape(node_object.value, shape)

            variable_dict[node_name] = VariableData(
                node_object.value_type(), value, domain
            )

        # remove data_loop and predictive_samples plates
        self._remove_plate(DATA_LOOP)

        # undo changes to target variables
        for node_name, old_state in old_observed_state.items():
            node_object = self.get_node_object(node_name)
            node_object.observed = old_state

        return BayesianNetworkDataset(variable_dict)

    def write_dot(self, dot_file_name: str) -> None:  # pragma: no cover
        """Save the model DAG to a 'dot' file."""
        nx.nx_pydot.write_dot(self.dag, dot_file_name)

    def write_net(self, net_file_name: str) -> None:
        """Save the Bayesian Network in HUGIN NET format."""

        assert all(
            isinstance(n, CategoricalNodeWithDirichletPrior)
            for n in self.get_node_dict().values()
        ), "write_net() not supported for models with non-CATEGORICAL nodes."

        with open(net_file_name, "w") as f:
            # write header
            f.write("net\n")
            f.write("{\n")
            f.write("\tname = {};\n".format(self.name))
            f.write("}\n")

            # write nodes
            for node_name, node_object in self.get_node_dict().items():
                f.write("node {}\n".format(node_name))
                f.write("{\n")
                if node_object.domain is None:
                    domain = range(node_object.domain_size)
                else:
                    domain = node_object.domain
                f.write(
                    "\tstates = ({});\n".format(
                        " ".join('"{}"'.format(w) for w in domain)
                    )
                )
                f.write("}\n")

            # write potentials (i.e., cpds)
            for node_name, node_object in self.get_node_dict().items():
                f.write(
                    "potential ( {} | {})\n".format(
                        node_name, " ".join(self.dag.predecessors(node_name))
                    )
                )
                f.write("{\n")
                _ = node_object.MAP_cpd()
                f.write(
                    "\tdata = {};\n".format(
                        hugin_potential_string(node_object.guide_MAP_cpd)
                    )
                )
                f.write("}\n")

    def overwrite_param_store(self) -> None:
        """Overwrite the parameters in the Pyro param store with own parameters"""
        for node_object in self.get_node_dict().values():
            node_object.overwrite_param_store()

    def validate(self, dataset: BayesianNetworkDataset = None) -> None:
        """
        Check that the graphical model has a valid structure.
        Optionally check that the graphical model and the dataset
        agree on the domain of categorical variables.

            - check that the graph is directed acyclic

            - check that the child nodes don't exist outside of the
              plates of their parents.

            - Given a dataset, check that domain outcomes match
              between the graphical model and the dataset
        """

        if not nx.is_directed_acyclic_graph(self.dag):
            raise ValueError("Graph is not a DAG")

        for node in self.dag.nodes:
            node_plates = self.get_node_object(node).plates
            for child in self.dag.successors(node):
                child_plates = self.get_node_object(child).plates
                if not set(node_plates).issubset(child_plates):
                    raise ValueError(
                        f"Plates of node {node} are not included in the plates of its child {child}"
                    )

        if dataset is not None:
            for model_var_name, model_node in self.get_node_dict().items():
                if (
                    model_node.value_type() == NodeValueType.CATEGORICAL
                    and model_var_name in dataset.variable_dict
                ):
                    if model_node.domain != dataset.discrete_domain(model_var_name):
                        raise ValueError(
                            f"Mismatch in domain of categorical variable: "
                            f"{model_node.domain} in model vs. "
                            f"{dataset.discrete_domain(model_var_name)} in dataset."
                        )

                    var_value = dataset[model_var_name]
                    if not torch.all(var_value < model_node.domain_size):
                        raise ValueError(
                            f"Values for node {model_node.name} must be < domain size ({model_node.domain_size}): {var_value}"
                        )

                    if not torch.all(0 <= var_value):
                        raise ValueError(
                            f"Values for node {model_node.name} must be >= 0: {var_value}"
                        )

    def _add_outer_plate(self, name: str) -> None:
        min_dim = min((p.dim for p in self.plate_dict.values()), default=0)
        self.plate_dict[name] = Plate(name, min_dim - 1)
        for node in self.dag.nodes:
            node_object = self.get_node_object(node)
            node_object.plates.append(name)

    def _remove_plate(self, name: str) -> None:
        self.plate_dict.pop(name)
        for node in self.dag.nodes:
            node_object = self.get_node_object(node)
            node_object.plates.remove(name)

    def to_yaml_encoding(self) -> Mapping[str, Any]:
        yaml_encoding = dict()
        yaml_encoding["encodingVersion"] = ENCODING_VERSION
        yaml_encoding["name"] = self.name
        yaml_encoding["device"] = {
            "type": self.device.type,
            "index": self.device.index,
        }
        yaml_encoding["plates"] = {k: v.dim for k, v in self.plate_dict.items()}
        yaml_encoding["nodes"] = dict()
        for node in nx.topological_sort(self.dag):
            node_object = self.get_node_object(node)
            yaml_encoding["nodes"][node] = node_object.to_yaml_encoding()
        return yaml_encoding

    @classmethod
    def from_yaml_encoding(cls, yaml_encoding: Mapping[str, Any]) -> "BayesianNetwork":
        if not yaml_encoding["encodingVersion"] == ENCODING_VERSION:
            raise ValueError(
                f"expected encoding version: {ENCODING_VERSION}, received {yaml_encoding['encodingVersion']}"
            )

        device = torch.device(
            yaml_encoding["device"]["type"], yaml_encoding["device"]["index"]
        )
        model = cls(yaml_encoding["name"], device=device)
        for p, d in yaml_encoding["plates"].items():
            model.add_plate(p, d)

        template_dag = nx.DiGraph()
        for n in yaml_encoding["nodes"]:
            template_dag.add_node(n)
            for p in yaml_encoding["nodes"][n]["parents"]:
                template_dag.add_edge(p, n)

        for n in nx.topological_sort(template_dag):
            node_spec = yaml_encoding["nodes"][n]
            node_cls = NODE_CLASS[node_spec["type"]]
            node_object = node_cls.from_yaml_encoding(n, model, node_spec)
            node_parents = [model.get_node_object(p) for p in node_spec["parents"]]
            model._add_node(node_object, node_parents)
        return model

    def _infer_plate_sizes(self, dataset: BayesianNetworkDataset) -> Dict[str, int]:
        """Infer plate sizes from the dataset"""
        plate_sizes = dict()
        for node_name, node_object in self.get_node_dict().items():
            relevant_plates = [self.plate_dict[p] for p in node_object.plates]
            relevant_plates.sort(key=lambda p: p.dim)
            data_dims = dataset[node_name].size()
            for plate, size in zip(relevant_plates, data_dims):
                if plate.name in plate_sizes and plate_sizes[plate.name] != size:
                    raise ValueError(
                        f"dataset doesn't assign a unique size to plate: {plate.name}"
                    )
                else:
                    plate_sizes[plate.name] = size
        return plate_sizes

    def to(self, device: torch.device) -> "BayesianNetwork":
        """
        Create a copy of the ``BayesianNetwork`` object with tensors
        copied to the specified device.  Return a shallow copy if the
        object is already using the specified device.

        :param device: torch device to be used for the tensors of the
            copied ``BayesianNetwork`` object.
        """
        if same_device(self.device, device):
            return copy.copy(self)
        else:
            yaml_encoding = self.to_yaml_encoding()
            yaml_encoding["device"] = {"type": device.type, "index": device.index}
            return BayesianNetwork.from_yaml_encoding(yaml_encoding)

    def cpu(self) -> "BayesianNetwork":
        """
        Return a copy of the ``BayesianNetwork`` object with tensors
        copied to CPU memory.  If the object is already using CPU
        memory a shallow copy is returned.
        """
        return self.to(torch.device("cpu", 0))

    def cuda(self, device: torch.device) -> "BayesianNetwork":
        """
        Return a copy of the ``BayesianNetwork`` object with tensors
        copied to the specified CUDA device.  If the object is already
        using the specified device a shallow copy is returned.

        :param device: torch cuda device to be used to store the
            tensors of the copied ``BayesianNetwork`` object.
        """
        if not device.type == "cuda":
            raise ValueError(f"did not receive cuda device as input: {device}")
        return self.to(device)

    def parse_evidence(self, evidence: str) -> Dict:
        """
        Parse a string with evidence on the random variables of the
        BayesianNetwork.

        :param evidence: string containing the evidence on random
            variables.  The string has the form of a comma separated
            list of equality constraints.  The evidence string should
            be well-formed and use values from the domain of the
            random variable.

            Examples:

                - "EV0016=True"

                - "EV0016=True,Interpretation=Pathogenic"

                - "Interpretation=Pathogenic, Dma.Score=1.0"

        :return: A dictionary mapping variables in the evidence to
                 their values.
        """
        evidence_dict = dict()
        for e in evidence.replace(" ", "").split(","):
            if not e:
                continue

            variable, value = e.split("=")
            if variable in evidence_dict:
                raise ValueError(
                    f"multiple equality constraints on {variable} in {evidence}"
                )
            node_object = self.get_node_object(variable)
            if node_object.value_type() == NodeValueType.CATEGORICAL:
                evidence_dict[variable] = node_object.domain.index(value)
            else:
                evidence_dict[variable] = float(value)

        return evidence_dict

    def parse_query(self, query: str) -> List:
        """
        Parse a string containing query on the random variables of the
        BayesianNetwork.

        :param query: string containing the query on the random
            variables.  The string has the form of a comma separated
            list of parts.  Each part is either an unconstrained or
            equality constrained random variable.  The query string
            should be well-formed.  Equality constraints are allowed
            only for categorical random variables.

            Examples:

                - "Interpretation"

                - "Interpretation=Pathogenic"

                - "Interpretation=Pathogenic, EV0016=True"

                - "Interpretation=Pathogenic, EV0016"

                - "Interpretation=Pathogenic, EV0016, Dma.Score"

        :return: A tuple of three lists.

                - list of equality constraints as (variable, value)
                  pairs

                - list of unconstrained categorical variables

                - lisf of unconstrained continuous variables
        """
        constraints = []
        unconstrained_discrete_variables = []
        unconstrained_continuous_variables = []
        for q in query.replace(" ", "").split(","):
            if not q:
                continue
            q_parts = q.split("=")
            if len(q_parts) == 1:
                node_object = self.get_node_object(q)
                if node_object.value_type() == NodeValueType.CATEGORICAL:
                    unconstrained_discrete_variables.append(q)
                else:
                    unconstrained_continuous_variables.append(q)
            elif len(q_parts) == 2:
                node_object = self.get_node_object(q_parts[0])
                constraints.append((q_parts[0], node_object.domain.index(q_parts[1])))
            else:
                raise ValueError(f"Invalid query {q} in {query}")
        return (
            constraints,
            unconstrained_discrete_variables,
            unconstrained_continuous_variables,
        )


def to_yaml(bn: BayesianNetwork, file: str) -> None:
    with open(file, "w") as fp:
        yaml.dump(bn.to_yaml_encoding(), fp)


def from_yaml(file: str) -> BayesianNetwork:
    with open(file, "r") as fp:
        yaml_encoding = yaml.safe_load(fp)
        return BayesianNetwork.from_yaml_encoding(yaml_encoding)



import torch


def hugin_potential_string(t: torch.Tensor) -> str:
    if t.ndim == 1:
        return "({})".format(" ".join(map(lambda x: "{:.5f}".format(x), map(float, t))))
    else:
        return "({})".format(" ".join(hugin_potential_string(i) for i in t))


def is_hugin_identifier(s: str) -> bool:
    # HUGIN identifiers are supposed to be valid 'C' language
    # identifiers.  This may not be the exact test for checking if a
    # string is a valid C identifier.  But we are assuming that users
    # are not trying to break the code.
    if s.isidentifier() and not s.startswith("_"):
        return True
    else:
        return False



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pearl', 'pearl.nodes', 'pearl.structure']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'h5py>=3.2.1,<4.0.0',
 'networkx>=2.4,<3.0',
 'numpy>=1.20.3,<2.0.0',
 'pyro-ppl>=1.4.0,<2.0.0',
 'torch>=1.6.0,<2.0.0']

setup_kwargs = {
    'name': 'pearl-pgm',
    'version': '0.2.0',
    'description': 'Library for declarative specification of directed graphical models and inference using Pyro.',
    'long_description': '# pearl-pgm\n\nLibrary for declarative specification of directed graphical models and inference using Pyro.\n',
    'author': 'Arun Nampally',
    'author_email': 'arun.nampally@invitae.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/invitae/pearl-pgm',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from contextlib import ExitStack
from typing import Any, List, Mapping, Optional, Sequence

import pyro
import pyro.distributions as distributions
import torch
from pyro.ops.indexing import Vindex

from pearl.common import NodeValueType, Plate, SamplingMode, same_device
from pearl.csi import TensorIndexMapping


class Node:
    """Node in a Bayesian Network"""

    def __init__(
        self,
        name: str,
        plates: List[str],
        parents: Sequence[Node],
        device: torch.device = torch.device("cpu"),
        prior_params: Mapping[str, torch.Tensor] = None,
        observed: bool = True,
    ) -> None:
        """
        :param name: name of the node (i.e., random variable)
        :param plates: list of plates in which the node is embedded
        :param parents: list of parent nodes
        :param device: cpu/cuda device to use for storing tensors
        :param prior_params: a mapping from param name to prior values.  See implementations for specifics of what
        prior params are accepted and their interpretation.
        :param observed: whether the node is observed or hidden
        """
        self.name = name
        self.device = device
        self.parents = parents

        self.prior_params = dict(
            self.get_default_prior_params(), **(prior_params or {})
        )

        self.validate_prior_params()
        self.observed = observed
        self.value = None  # sampled value; can have enumeration + batch dimensions
        self.observed_value = None  # observed value; can have batch dimensions
        self.plates = [] if plates is None else list(plates)

    def set_observed_value(
        self, value: Optional[torch.Tensor]
    ) -> None:  # pragma: no cover
        """Set the value of the node."""
        if value is not None:
            assert same_device(value.device, self.device)

        self.observed_value = value

    def sample_model_cpd(self) -> torch.Tensor:  # pragma: no cover
        """Sample the model CPD from its prior."""
        raise NotImplementedError

    def sample_guide_cpd(self) -> torch.Tensor:  # pragma: no cover
        """Sample the guide CPD from its prior."""
        raise NotImplementedError

    def MAP_cpd(self) -> torch.Tensor:  # pragma: no cover
        """Record the MAP of the posterior on the CPD of the node and return it."""
        raise NotImplementedError

    def sample(
        self,
        relevant_plates: List[Plate],
        sampling_mode: SamplingMode,
        *parent_values: torch.Tensor,
    ) -> None:
        """
        First obtain the node-specific distribution given any parent values.
        If this node is hidden then sample from the distribution.
        If this node is observed but we are currently guide mode then skip any sampling.
        If this node is observed and we are not currently in guide mode then create a sample
         site for the distributuion from with observations set.
        """

        sample_dist = self.get_sample_dist(sampling_mode, *parent_values)

        with ExitStack() as stack:
            idxs = tuple(
                stack.enter_context(plate.pyro_plate) for plate in relevant_plates
            )
            if not self.observed:
                # unobserved nodes are always sampled
                self.value = pyro.sample(self.name, sample_dist, obs=None).float()
            else:
                obs = self.subsample_observations(relevant_plates, idxs)
                # observed nodes are sampled except in guide
                if sampling_mode == SamplingMode.GUIDE:
                    self.value = obs
                else:
                    self.value = pyro.sample(self.name, sample_dist, obs=obs).float()

    def get_sample_dist(
        self, sampling_mode: SamplingMode, *parent_values: torch.Tensor
    ) -> distributions.Distribution:
        """Given any parent values, return the Pyro distribution from which the node's values should be sampled."""
        raise NotImplementedError

    def subsample_observations(
        self, relevant_plates: List[Plate], idxs: Sequence[torch.tensor]
    ):
        """Broadcast the 1d indices returned by plates so that data has the right batch shape

        Let -m be the minimum dimension (most leftmost/negative) of the relevant plates for this node. Then the
        pyro.sample site for this node is expecting observations that have dimensions from -m to -1. Ultimately,
        the left-most plate during training and prediction, data_loop and predictive_samples, respectively,
        are added to every node. Therefore all pyro.sample sites in pearl will expect observations with number of
        dimensions equal to the total number of plates.

        Observations in the user-created dataset for each node only contain dimensions for the plates relevant to
        that node (data_loop plus any user-defined plates). This method reshapes the self.observed_value tensor to
        fill out those missing dimensions (along with actually subsampling). As an example, a variable that lies on
        plates with dim index -4 and -2 will be have self.observed_value as a 2-dimensional tensor with size [x,
        y]. This method will reshape it to a tensor with size [x, 1, y, 1].

        This is done by reshaping each of the indices coming from the plate contexts in relevant_plates. Each of
        those indices idx for a plate at dimension -d need to get singleton dimensions tacked on from [-d+1, ...,
        -1] so that idx has shape [plate_size, 1, ..., 1], with d dimensions. Once the indices are broadcaseted,
        Vindex takes care of everything else, including dealing with enumerated dimensions.
        """

        # if dim=-N add abs(N) - 1 singleton dimensions to the right.
        dims = tuple(plate.dim for plate in relevant_plates)
        broadcasted_idxs = tuple(
            idx.view((-1,) + (1,) * (abs(dim) - 1)) for idx, dim in zip(idxs, dims)
        )
        obs = Vindex(self.observed_value)[broadcasted_idxs]

        return obs

    def overwrite_param_store(self) -> None:  # pragma: no cover
        """Overwrite node parameters in the Pyro param store with current node attributes."""
        raise NotImplementedError

    def value_type(self) -> NodeValueType:  # pragma: no cover
        """
        :return value type of node, whether categorical or continuous.
        """
        raise NotImplementedError

    def validate_prior_params(self) -> None:
        """Ensure that the prior params that are passed in are valid. In general, each class's method should call
        that of its parent class as well.
        """
        for tensor in self.prior_params.values():
            assert same_device(tensor.device, self.device)

    def get_default_prior_params(self) -> Mapping[str, torch.Tensor]:
        """
        :return str --> torch.Tensor dictionary of configured default params.
        """
        return dict()

    def dist(self, tensor) -> distributions.Distribution:  # pragma: no cover
        """
        :return the distribution used for this class.
        """
        raise NotImplementedError

    def __eq__(self, other) -> bool:  # pragma: no cover
        """
        Test whether the node is equal to the 'other' node.
        """
        raise NotImplementedError

    def to_yaml_encoding(self) -> Mapping[str, Any]:
        yaml_encoding = dict()
        yaml_encoding["type"] = type(self).__name__
        yaml_encoding["plates"] = list(self.plates)
        yaml_encoding["parents"] = [p.name for p in self.parents]
        yaml_encoding["observed"] = self.observed
        return yaml_encoding

    @classmethod
    def process_yaml_spec(
        cls, bayesnet, yaml_encoding: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        kwargs = dict()
        kwargs["plates"] = yaml_encoding["plates"]
        kwargs["parents"] = [
            bayesnet.get_node_object(p) for p in yaml_encoding["parents"]
        ]
        kwargs["device"] = bayesnet.device
        if "observed" in yaml_encoding:
            kwargs["observed"] = yaml_encoding["observed"]
        return kwargs

    @classmethod
    def from_yaml_encoding(
        cls, name, bayesnet, yaml_encoding: Mapping[str, Any]
    ) -> "Node":
        kwargs = cls.process_yaml_spec(bayesnet, yaml_encoding)
        kwargs["name"] = name
        return cls(**kwargs)


class NodeWithCategoricalParents(Node):
    def __init__(
        self,
        name: str,
        plates: List[str],
        parents: Sequence[Node],
        **kwargs,
    ):
        if parents is None:
            parents = []
        self.parent_domain_sizes = torch.Size([p.domain_size for p in parents])
        super().__init__(name, plates, parents, **kwargs)

        assert all([p.value_type() == NodeValueType.CATEGORICAL for p in parents])

        # context specific independence
        self.model_csi = self._init_csi(self.parent_domain_sizes)
        self.guide_csi = self._init_csi(self.parent_domain_sizes)

    def _init_csi(
        self, size: torch.Size
    ) -> Optional[TensorIndexMapping]:  # pragma: no cover

        """Create initial structure for storing context specific independence."""
        if len(size) > 0:
            return TensorIndexMapping(size, self.device)
        else:
            return None

    def _add_csi_rule(self, csi, partial_mapping):
        assert csi is not None
        assert len(partial_mapping) <= len(self.parent_domain_sizes)
        indices = tuple(
            partial_mapping.get(d, slice(None, None, None))
            for d in range(len(self.parent_domain_sizes))
        )
        canonical_index = tuple(
            partial_mapping.get(d, 0) for d in range(len(self.parent_domain_sizes))
        )
        csi.add_mapping(indices, canonical_index)

    def model_add_csi_rule(self, partial_mapping: Mapping[int, int]) -> None:
        """Add context specific independence rule to model CPD.

        :param partial_mapping: context for the csi rule
        """

        self._add_csi_rule(self.model_csi, partial_mapping)

    def guide_add_csi_rule(self, partial_mapping: Mapping[int, int]) -> None:
        """Add context specific independence rule to guide CPD.

        :param partial_mapping: context for the csi rule
        """

        self._add_csi_rule(self.guide_csi, partial_mapping)

    def _index_cpd(
        self, cpd: torch.Tensor, *parent_values: torch.Tensor
    ) -> List[torch.Tensor]:
        """
        Index the conditional probability distribution of the node using parent values and return it.

        :param cpd: tensor representing the CPD of the node
        :param parent_values: tensors containing values of the parents of the node
        """
        assert len(parent_values) == len(self.parent_domain_sizes)
        if len(parent_values) > 0:
            t_parent_values = tuple(p.long() for p in parent_values)
            return Vindex(cpd)[t_parent_values]
        else:
            return cpd

    def get_sample_dist(
        self, sampling_mode: SamplingMode, *parent_values: torch.Tensor
    ) -> distributions.Distribution:
        """
        Get the CPD for this node based on its parents' values
        and return the corresponding distribution function.
        """

        # sample the appropriate cpd
        if sampling_mode == SamplingMode.MODEL:
            cpd = self._index_cpd(self.sample_model_cpd(), *parent_values)
        elif sampling_mode == SamplingMode.GUIDE:
            cpd = self._index_cpd(self.sample_guide_cpd(), *parent_values)
        elif sampling_mode == SamplingMode.MAP_POSTERIOR:
            cpd = self._index_cpd(self.MAP_cpd(), *parent_values)
        else:
            raise ValueError(f"unrecognized sampling mode- {sampling_mode}")

        return self.dist(cpd)
